from __future__ import annotations

import dataclasses
import typing as _t
from dataclasses import dataclass, field
from enum import Enum

from syntax_diagrams.element import Element
from syntax_diagrams.measure import SimpleTextMeasure, TextMeasure
from syntax_diagrams.resolver import HrefResolver

__all__ = [
    "LoadingError",
    "EndClass",
    "TextRenderSettings",
    "render_text",
    "ArrowStyle",
    "SvgRenderSettings",
    "DEFAULT_CSS",
    "render_svg",
]

T = _t.TypeVar("T")
U = _t.TypeVar("U")


class LoadingError(ValueError):
    """
    Indicates incorrect diagram value.

    """


class EndClass(Enum):
    """
    Controls how diagram start and end look like.

    """

    SIMPLE = "SIMPLE"
    """
    A simple ``T``-shaped ending.

    .. syntax:diagram::
        :end-class: simple

        Simple

    """

    COMPLEX = "COMPLEX"
    """
    A ``T``-shaped ending with vertical line doubled.

    .. syntax:diagram::
        :end-class: complex

        Complex

    """


@dataclass(frozen=True)
class TextRenderSettings:
    """
    Settings for text rendering engine.

    """

    padding: tuple[int, int, int, int] = (0, 0, 0, 0)
    """
    Array of four positive integers denoting top, right, bottom and left
    padding between the diagram and its container.

    """

    vertical_choice_separation_outer: int = 1
    """
    Vertical space between nodes in a :func:`choice` block,
    if it contains another choice block.

    """

    vertical_choice_separation: int = 1
    """
    Vertical space between nodes in a :func:`choice` block.

    """

    vertical_seq_separation_outer: int = 1
    """
    Vertical space between nodes in a :func:`stack` block,
    if it appears outside of any choice block.

    """

    vertical_seq_separation: int = 1
    """
    Vertical space between nodes in a :func:`stack` block.

    """

    horizontal_seq_separation: int = 2
    """
    Horizontal space between adjacent nodes.

    """

    group_vertical_padding: int = 1
    """
    Vertical padding inside of group elements.

    """

    group_horizontal_padding: int = 2
    """
    Horizontal padding inside of group elements.

    """

    group_vertical_margin: int = 0
    """
    Vertical margin outside of group elements.

    """

    group_horizontal_margin: int = 2
    """
    Horizontal margin outside of group elements.

    """

    group_text_vertical_offset: int = -1
    """
    Offset from group rectangle to its heading.

    """

    group_text_horizontal_offset: int = 2
    """
    Offset from group rectangle to its heading.

    """

    end_class: EndClass = EndClass.COMPLEX
    """
    Controls how diagram start and end look like.

    """

    max_width: int = 80
    """
    Max width after which a sequence will be wrapped. This option is used to
    automatically convert sequences to stacks. Note that this is a suggestive
    option, there is no guarantee that the diagram will
    fit to its ``max_width``.

    """

    reverse: bool = False
    """
    If enabled, diagram is rendered right-to-left.

    """


@_t.overload
def render_text(
    element: Element[T],
    /,
    *,
    max_width: int | None = None,
    settings: TextRenderSettings | None = None,
    reverse: bool | None = None,
    href_resolver: HrefResolver[T] | None = None,
    convert_resolver_data: None = None,
) -> str: ...


@_t.overload
def render_text(
    element: Element[U],
    /,
    *,
    max_width: int | None = None,
    settings: TextRenderSettings | None = None,
    reverse: bool | None = None,
    href_resolver: HrefResolver[T] | None = None,
    convert_resolver_data: _t.Callable[[U | None], T | None] | None,
) -> str: ...


def render_text(
    node: Element[T],
    /,
    *,
    max_width: int | None = None,
    settings: TextRenderSettings | None = None,
    reverse: bool | None = None,
    href_resolver: HrefResolver[T] | None = None,
    convert_resolver_data: _t.Callable[[_t.Any | None], T | None] | None = None,
    _dump_debug_data: bool = False,
) -> str:
    """
    Render diagram as an SVG.

    :param node:
        root diagram node.
    :param max_width:
        if given, overrides `~SvgRenderSettings.max_width`
        from ``settings``.
    :param settings:
        rendering settings.
    :param reverse:
        if given, overrides `~SvgRenderSettings.reverse`
        from ``settings``.
    :param href_resolver:
        hyperlink resolver (may be used to transform text).
    :param convert_resolver_data:
        a custom function to parse ``resolver_data``, used when loading elements
        from JSON or other untrusted source.
    :return:
        a string with a rendered diagram.
    :throws:
        `LoadingError`

    """

    if convert_resolver_data is None:
        convert_resolver_data = lambda x: x
    if settings is None:
        settings = TextRenderSettings()
    if href_resolver is None:
        href_resolver = HrefResolver()

    # Prevent circular import.
    from syntax_diagrams._impl.load import load as _load
    from syntax_diagrams._impl.render.text import render_text as _render_text

    return _render_text(
        _load(node, convert_resolver_data),
        max_width=max_width,
        settings=settings,
        reverse=reverse,
        href_resolver=href_resolver,
        dump_debug_data=_dump_debug_data,
    )


DEFAULT_CSS = {
    "path": {
        "stroke-width": "1.5",
        "stroke": "black",
        "fill": "none",
        "stroke-linecap": "round",
    },
    ".arrow": {
        "stroke": "none",
        "fill": "black",
    },
    ".arrow.arrow-barb": {
        "stroke": "black",
        "fill": "none",
    },
    "a": {
        "text-decoration": "none",
    },
    "rect": {
        "stroke-width": "1.5",
        "stroke": "black",
        "fill": "none",
    },
    "g.comment rect": {
        "stroke": "none",
    },
    "text": {
        "font-family": "Consolas, Menlo, monospace",
        "text-anchor": "middle",
        # "dominant-baseline": "central",
        "font-weight": "bold",
    },
    "tspan": {
        "font-family": "Consolas, Menlo, monospace",
    },
    ".escape": {
        "fill": "brown",
    },
    ".group text": {
        "text-anchor": "start",
        "opacity": "0.5",
        "font-weight": "normal",
        "font-style": "italic",
        # "dominant-baseline": "alphabetic",
    },
    ".group rect": {
        "opacity": "0.3",
    },
    "g.comment text": {
        "font-weight": "normal",
        "font-style": "italic",
    },
    ".dbg-content, .dbg-display, .dbg-content-main, .dbg-padding, .dbg-margin, .dbg-position, .dbg-ridge-line": {
        "stroke": "none",
    },
}


# Measured for Menlo, 14px. Consolas is a bit narrower, but not too much.
_DEFAULT_TEXT_MEASURE_FACTORY = lambda: SimpleTextMeasure(
    character_advance=8.44,
    wide_character_advance=14.34,
    font_size=14,
    line_height=14 * 1.1,
    ascent=12,
)


class ArrowStyle(Enum):
    """
    Arrow shapes for SVG diagrams.

    """

    NONE = "NONE"
    """
    Arrows are not rendered.

    .. syntax:diagram::
        :svg-arrow-style: none
        :svg-padding: 10 10 10 10

        optional: "node"
        skip: true

    """

    TRIANGLE = "TRIANGLE"
    """
    Simple triangle arrows.

    .. syntax:diagram::
        :svg-arrow-style: triangle
        :svg-padding: 10 10 10 10

        optional: "node"
        skip: true

    """

    STEALTH = "STEALTH"
    """
    Pointier triangle arrows.

    .. syntax:diagram::
        :svg-arrow-style: stealth
        :svg-padding: 10 10 10 10

        optional: "node"
        skip: true

    """

    BARB = "BARB"
    """
    An error that consists of two lines.

    .. syntax:diagram::
        :svg-arrow-style: barb
        :svg-padding: 10 10 10 10

        optional: "node"
        skip: true

    """

    HARPOON = "HARPOON"
    """
    Half of a triangle arrow.

    .. syntax:diagram::
        :svg-arrow-style: harpoon
        :svg-padding: 10 10 10 10

        optional: "node"
        skip: true

    """

    HARPOON_UP = "HARPOON_UP"
    """
    Another half of a triangle arrow.

    .. syntax:diagram::
        :svg-arrow-style: harpoon-up
        :svg-padding: 10 10 10 10

        optional: "node"
        skip: true

    """


@dataclass(frozen=True)
class SvgRenderSettings:
    """
    Settings for SVG rendering engine.

    """

    padding: tuple[int, int, int, int] = (1, 1, 1, 1)
    """
    Array of four positive integers denoting top, right, bottom and left
    padding between the diagram and its container.

    """

    title: str | None = None
    """
    Title text that will be added to ``<title>`` element and ``aria-label`` attribute.

    """

    description: str | None = None
    """
    Title text that will be added to ``<desc>`` element.

    """

    vertical_choice_separation_outer: int = 9
    """
    Vertical space between nodes in a :func:`choice` block,
    if it contains another choice block.

    """

    vertical_choice_separation: int = 9
    """
    Vertical space between nodes in a :func:`choice` block.

    """

    vertical_seq_separation_outer: int = 18
    """
    Vertical space between nodes in a :func:`stack` block,
    if it appears outside of any choice block.

    """

    vertical_seq_separation: int = 18
    """
    Vertical space between nodes in a :func:`stack` block.

    """

    horizontal_seq_separation: int = 10
    """
    Horizontal space between adjacent nodes.

    """

    end_class: EndClass = EndClass.COMPLEX
    """
    Controls how diagram start and end look like.

    """

    arrow_style: ArrowStyle = ArrowStyle.NONE
    """
    Style of arrows to draw.

    """

    arrow_length: int = 10
    """
    Length of an arrow along its line.

    """

    arrow_cross_length: int = 4
    """
    Length of an arrow across its line.

    """

    max_width: int = 600
    """
    Max width after which a sequence will be wrapped. This option is used to
    automatically convert sequences to stacks. Note that this is a suggestive
    option, there is no guarantee that the diagram will
    fit to its ``max_width``.

    """

    arc_radius: int = 10
    """
    Arc radius of railroads. 10px by default.

    """

    arc_margin: int = 5
    """
    Margin around arcs.

    """

    terminal_text_measure: TextMeasure = dataclasses.field(
        default_factory=_DEFAULT_TEXT_MEASURE_FACTORY
    )
    """
    Service used to measure size of text inside of text nodes.

    By default, set up to match default CSS settings.

    """

    terminal_horizontal_padding: int = 10
    """
    Horizontal padding around text in terminal nodes.

    """

    terminal_vertical_padding: int = 3
    """
    Vertical padding around text in terminal nodes.

    """

    terminal_radius: int = 10
    """
    Border radius in terminal nodes.

    """

    non_terminal_text_measure: TextMeasure = dataclasses.field(
        default_factory=_DEFAULT_TEXT_MEASURE_FACTORY
    )
    """
    Service used to measure size of text inside of text nodes.

    By default, set up to match default CSS settings.

    """

    non_terminal_horizontal_padding: int = 10
    """
    Horizontal padding around text in non-terminal nodes.

    """

    non_terminal_vertical_padding: int = 3
    """
    Vertical padding around text in non-terminal nodes.

    """

    non_terminal_radius: int = 0
    """
    Border radius in non-terminal nodes.

    """

    comment_text_measure: TextMeasure = dataclasses.field(
        default_factory=_DEFAULT_TEXT_MEASURE_FACTORY
    )
    """
    Service used to measure size of text inside of text nodes.

    By default, set up to match default CSS settings.

    """

    comment_horizontal_padding: int = 3
    """
    Horizontal padding around text in comment nodes.

    """

    comment_vertical_padding: int = 3
    """
    Vertical padding around text in comment nodes.

    """

    comment_radius: int = 0
    """
    Border radius in comment nodes.

    """

    group_text_measure: TextMeasure = dataclasses.field(
        default_factory=_DEFAULT_TEXT_MEASURE_FACTORY
    )
    """
    Service used to measure size of text of groups.

    By default, set up to match default CSS settings.

    """

    group_vertical_padding: int = 10
    """
    Vertical padding inside of group elements.

    """

    group_horizontal_padding: int = 10
    """
    Horizontal padding inside of group elements.

    """

    group_vertical_margin: int = 5
    """
    Vertical margin outside of group elements.

    """

    group_horizontal_margin: int = 10
    """
    Horizontal margin outside of group elements.

    """

    group_radius: int = 0
    """
    Border radius in groups.

    """

    group_text_vertical_offset: int = 0
    """
    Offset from group rectangle to its heading.

    """

    group_text_horizontal_offset: int = 10
    """
    Offset from group rectangle to its heading.

    """

    css_class: str = ""
    """
    Class attribute for the ``<svg>`` element.

    """

    css_style: dict[str, dict[str, str]] | str | None = field(
        default_factory=lambda: DEFAULT_CSS
    )
    """
    CSS style that should be embedded into the diagram.

    """

    reverse: bool = False
    """
    If enabled, diagram is rendered right-to-left.

    """


@_t.overload
def render_svg(
    element: Element[T],
    /,
    *,
    max_width: int | None = None,
    settings: SvgRenderSettings | None = None,
    reverse: bool | None = None,
    href_resolver: HrefResolver[T] | None = None,
    convert_resolver_data: None = None,
) -> str: ...


@_t.overload
def render_svg(
    element: Element[U],
    /,
    *,
    max_width: int | None = None,
    settings: SvgRenderSettings | None = None,
    reverse: bool | None = None,
    href_resolver: HrefResolver[T] | None = None,
    convert_resolver_data: _t.Callable[[U | None], T | None] | None,
) -> str: ...


def render_svg(
    element: Element[T],
    /,
    *,
    max_width: int | None = None,
    settings: SvgRenderSettings | None = None,
    reverse: bool | None = None,
    href_resolver: HrefResolver[T] | None = None,
    convert_resolver_data: _t.Callable[[T | None], T | None] | None = None,
    _dump_debug_data: bool = False,
) -> str:
    """
    Render diagram as an SVG.

    :param node:
        root diagram node.
    :param max_width:
        if given, overrides `~SvgRenderSettings.max_width`
        from ``settings``.
    :param settings:
        rendering settings.
    :param reverse:
        if given, overrides `~SvgRenderSettings.reverse`
        from ``settings``.
    :param href_resolver:
        hyperlink resolver (may be used to transform text).
    :param convert_resolver_data:
        a custom function to parse ``resolver_data``, used when loading elements
        from JSON or other untrusted source.
    :return:
        a string with a rendered diagram.
    :throws:
        `LoadingError`

    """

    if convert_resolver_data is None:
        convert_resolver_data = lambda x: x
    if settings is None:
        settings = SvgRenderSettings()
    if href_resolver is None:
        href_resolver = HrefResolver()

    # Prevent circular import.
    from syntax_diagrams._impl.load import load as _load
    from syntax_diagrams._impl.render.svg import render_svg as _render_svg

    return _render_svg(
        _load(element, convert_resolver_data),
        max_width=max_width,
        settings=settings,
        reverse=reverse,
        href_resolver=href_resolver,
        dump_debug_data=_dump_debug_data,
    )



import pytest

import syntax_diagrams as rr
from syntax_diagrams._impl.load import load
from syntax_diagrams._impl.tree.node import Node
from syntax_diagrams._impl.tree.sequence import Sequence
from syntax_diagrams._impl.tree.skip import Skip


def test_str():
    a = rr.terminal("A")
    b = rr.terminal("B")
    c = rr.terminal("C")

    assert str(load(rr.terminal("TOKEN"), lambda x: x)) == "TOKEN"
    assert str(load(rr.terminal("{"), lambda x: x)) == "'{'"
    assert str(load(rr.sequence(a, b, c), lambda x: x)) == "A B C"
    assert str(load(rr.sequence(a, rr.choice(b, c)), lambda x: x)) == "A (B | C)"
    assert str(load(rr.choice(a, rr.sequence(b, c)), lambda x: x)) == "A | B C"
    assert str(load(rr.sequence(a, rr.optional(b)), lambda x: x)) == "A B?"
    assert str(load(rr.choice(a, rr.optional(b)), lambda x: x)) == "(A | B)?"
    assert str(load(rr.choice(rr.optional(a), b), lambda x: x)) == "(A | B)?"
    assert str(load(rr.choice(rr.skip(), a, b), lambda x: x)) == "(A | B)?"
    assert str(load(rr.sequence(a, rr.sequence(b, c)), lambda x: x)) == "A B C"
    assert str(load(rr.choice(a, rr.choice(b, c)), lambda x: x)) == "A | B | C"
    assert str(load(rr.one_or_more(a), lambda x: x)) == "A+"
    assert str(load(rr.one_or_more(rr.sequence(a, b, c)), lambda x: x)) == "(A B C)+"
    assert (
        str(load(rr.one_or_more(rr.sequence(a, b), repeat=c), lambda x: x))
        == "A B (C A B)*"
    )
    assert (
        str(load(rr.one_or_more(rr.choice(a, b), repeat=c), lambda x: x))
        == "(A | B) (C (A | B))*"
    )
    assert str(load(rr.zero_or_more(a), lambda x: x)) == "A*"
    assert str(load(rr.zero_or_more(rr.sequence(a, b, c)), lambda x: x)) == "(A B C)*"
    assert (
        str(load(rr.zero_or_more(rr.sequence(a, b), repeat=c), lambda x: x))
        == "(A B (C A B)*)?"
    )


def test_seq_creation():
    a = rr.terminal("node A")
    b = rr.terminal("node B")
    c = rr.terminal("node C")

    assert isinstance(load(rr.sequence(), lambda x: x), Skip)
    assert isinstance(load(rr.sequence(a), lambda x: x), Node)
    assert isinstance(load(rr.sequence(a, b, c), lambda x: x), Sequence)

    with pytest.raises(ValueError):
        load(rr.sequence(a, b, linebreaks=[]), lambda x: x)

    with pytest.raises(ValueError):
        load(
            rr.sequence(a, b, linebreaks=[rr.LineBreak.SOFT, rr.LineBreak.SOFT]),
            lambda x: x,
        )



from __future__ import annotations

import math
import typing as _t
from functools import cached_property

from syntax_diagrams._impl.render import (
    ConnectionDirection,
    ConnectionType,
    LayoutContext,
    LayoutSettings,
    Render,
    RenderContext,
)
from syntax_diagrams._impl.ridge_line import (
    RidgeLine,
    merge_ridge_lines,
    reverse_ridge_line,
)
from syntax_diagrams._impl.tree import Element
from syntax_diagrams._impl.tree.barrier import Barrier
from syntax_diagrams._impl.tree.skip import Skip
from syntax_diagrams._impl.vec import Vec

T = _t.TypeVar("T")


class OneOrMore(Element[T], _t.Generic[T]):
    _item: Element[T]
    _repeat: Element[T]
    _repeat_top: bool

    # Layout info
    repeat_pos: list[int]

    def __new__(
        cls,
        item: Element[T],
        repeat: Element[T] | None = None,
        repeat_top: bool = False,
    ) -> Element[T]:
        repeat = repeat or Skip()

        if isinstance(item, Skip) and isinstance(repeat, Skip):
            return item

        self = super().__new__(cls)

        self._item = Barrier(item)
        self._repeat = repeat
        self._repeat_top = repeat_top  # TODO

        return self

    @cached_property
    def precedence(self) -> int:
        # OneOrMore without repeat renders as an unary operator
        return 3 if isinstance(self._repeat, Skip) else 2

    @cached_property
    def contains_choices(self) -> bool:
        return self._item.contains_choices or self._repeat.contains_choices

    def _calculate_content_layout(
        self, settings: LayoutSettings[T], context: LayoutContext
    ):
        assert self.context

        # Whether it's safe to place repeat arcs at the connection points,
        # or we need to shift them inwards.
        self._need_shift_start_arc = not (
            context.start_top_is_clear and context.start_bottom_is_clear
        )
        self._need_shift_end_arc = not (
            context.end_top_is_clear and context.end_bottom_is_clear
        )

        context = self._isolate(self._need_shift_start_arc, self._need_shift_end_arc)

        if context.is_outer:
            self._vertical_choice_separation = settings.vertical_choice_separation_outer
        else:
            self._vertical_choice_separation = settings.vertical_choice_separation

        arc_radius = math.ceil(settings.arc_radius)
        arc_size = settings.arc_margin + arc_radius

        self._start_arc_size = context.start_connection.arc_size(settings)
        self._additional_start_padding = (
            arc_size
            if self._need_shift_start_arc
            else max(arc_size, self._start_arc_size) - self._start_arc_size
        )
        self._end_arc_size = context.end_connection.arc_size(settings)
        self._additional_end_padding = (
            arc_size
            if self._need_shift_end_arc
            else max(arc_size, self._end_arc_size) - self._end_arc_size
        )

        item_context = LayoutContext(
            width=max(
                0,
                context.width
                - self._additional_start_padding
                - self._additional_end_padding,
            ),
            is_outer=False,
            start_connection=context.start_connection,
            start_direction=context.start_direction,
            end_connection=context.end_connection,
            end_direction=context.end_direction,
            allow_shrinking_stacks=context.allow_shrinking_stacks,
        )
        self._item.calculate_layout(settings, item_context)

        repeat_context = LayoutContext(
            width=max(
                0,
                context.width
                - self._additional_start_padding
                - self._additional_end_padding,
            ),
            is_outer=False,
            start_connection=ConnectionType.STACK,
            start_top_is_clear=True,
            start_bottom_is_clear=context.end_bottom_is_clear
            and (
                self._need_shift_end_arc
                or context.end_direction is not ConnectionDirection.DOWN
            ),
            start_direction=ConnectionDirection.UP,
            end_connection=ConnectionType.STACK,
            end_top_is_clear=True,
            end_bottom_is_clear=context.start_bottom_is_clear
            and (
                self._need_shift_start_arc
                or context.start_direction is not ConnectionDirection.DOWN
            ),
            end_direction=ConnectionDirection.UP,
            allow_shrinking_stacks=False,
        )
        self._repeat.calculate_layout(settings, repeat_context)

        width = max(
            self._item.width
            + self._additional_start_padding
            + self._additional_end_padding,
            self._repeat.width
            - 2 * arc_size
            + self._start_arc_size
            + self._end_arc_size
            + self._additional_start_padding
            + self._additional_end_padding,
        )
        display_width = max(
            self._item.display_width
            + self._additional_start_padding
            + self._additional_end_padding,
            self._repeat.display_width
            - 2 * arc_size
            + self._start_arc_size
            + self._end_arc_size
            + self._additional_start_padding
            + self._additional_end_padding,
        )

        self._center_offset = (
            self._start_arc_size
            + self._additional_start_padding
            + width
            - self._end_arc_size
            - self._additional_end_padding
        ) // 2

        self._repeat_content_width_l = self._repeat.content_width // 2
        self._repeat_content_width_r = (
            self._repeat.content_width - self._repeat_content_width_l
        )

        self.start_padding = min(
            self._item.start_padding + self._additional_start_padding,
            self._center_offset - self._repeat_content_width_l,
        )
        self.start_margin = max(
            0,
            self.start_padding
            - min(
                (
                    self._start_arc_size + arc_radius
                    if self._need_shift_start_arc
                    else (
                        self._start_arc_size
                        + self._additional_start_padding
                        - settings.arc_margin
                    )
                )
                - settings.arc_margin
                - arc_radius,
                self._item.start_padding
                + self._additional_start_padding
                - self._item.start_margin,
                self._center_offset
                - self._repeat_content_width_l
                - self._repeat.start_margin,
            ),
        )
        self.end_padding = max(
            0,
            width
            - max(
                self._item.start_padding
                + self._additional_start_padding
                + self._item.content_width,
                self._center_offset + self._repeat_content_width_r,
            ),
        )
        self.end_margin = max(
            0,
            max(
                self._item.start_padding
                + self._additional_start_padding
                + self._item.content_width
                + self._item.end_margin,
                self._center_offset
                + self._repeat_content_width_r
                + self._repeat.end_margin,
            )
            - (width - self.end_padding),
            self.end_padding
            - (
                self._end_arc_size + arc_radius
                if self._need_shift_end_arc
                else (
                    self._end_arc_size
                    + self._additional_end_padding
                    - settings.arc_margin
                )
            )
            + arc_radius
            + settings.arc_margin,
        )

        self.display_width = display_width
        self.content_width = max(0, width - self.start_padding - self.end_padding)

        self.up = self._item.up
        self.height = self._item.height
        self.down = (
            self._item.down
            + self._vertical_choice_separation
            + self._repeat.up
            + self._repeat.height
            + self._repeat.down
        )

    def _render_content(self, render: Render[T], context: RenderContext):
        arc_radius = math.ceil(render.settings.arc_radius)
        arc_size = render.settings.arc_margin + arc_radius

        assert self.context
        if self._need_shift_start_arc:
            (
                (render)
                .line(context.start_connection_pos, context.reverse)
                .segment_abs(context.start_connection_pos.x + context.dir * arc_size)
            )
        if self._need_shift_end_arc:
            (
                (render)
                .line(context.end_connection_pos, context.reverse)
                .segment_abs(context.end_connection_pos.x - context.dir * arc_size)
            )

        item_context = RenderContext(
            pos=context.pos
            + Vec(
                context.dir * self._additional_start_padding,
                0,
            ),
            reverse=context.reverse,
            start_connection_pos=(
                context.start_connection_pos
                + Vec(
                    (context.dir * arc_size if self._need_shift_start_arc else 0),
                    0,
                )
            ),
            end_connection_pos=(
                context.end_connection_pos
                - Vec(
                    (context.dir * arc_size if self._need_shift_end_arc else 0),
                    0,
                )
            ),
        )

        self._item.render(render, item_context)

        repeat_start_connection_pos = context.pos + Vec(
            context.dir
            * (
                self._start_arc_size + arc_radius
                if self._need_shift_start_arc
                else (
                    self._start_arc_size
                    + self._additional_start_padding
                    - render.settings.arc_margin
                )
            ),
            0,
        )
        repeat_end_connection_pos = context.pos + Vec(
            context.dir
            * (
                self.width
                - (
                    self._end_arc_size + arc_radius
                    if self._need_shift_end_arc
                    else (
                        self._end_arc_size
                        + self._additional_end_padding
                        - render.settings.arc_margin
                    )
                )
            ),
            self._item.height,
        )

        center = context.pos + Vec(context.dir * self._center_offset, 0)

        repeat_context = RenderContext(
            pos=(
                center
                + Vec(
                    context.dir
                    * (self._repeat.start_padding + self._repeat_content_width_r),
                    self._item.height
                    + self._item.down
                    + self._vertical_choice_separation
                    + self._repeat.up,
                )
            ),
            start_connection_pos=repeat_end_connection_pos,
            end_connection_pos=repeat_start_connection_pos,
            reverse=not context.reverse,
        )
        self._repeat.render(render, repeat_context)

        render.debug_pos(repeat_start_connection_pos)
        render.debug_pos(repeat_end_connection_pos)
        render.debug_pos(center)

    def _calculate_top_ridge_line(self) -> RidgeLine:
        ridge_line = self._item.top_ridge_line
        if self._additional_start_padding > 0:
            return ridge_line + Vec(self._additional_start_padding, 0)
        else:
            return ridge_line

    def _calculate_bottom_ridge_line(self) -> RidgeLine:
        assert self.settings

        arc_radius = math.ceil(self.settings.arc_radius)
        repeat_start_connection_pos = (
            self._start_arc_size
            if self._need_shift_start_arc
            else (
                self._start_arc_size
                + self._additional_start_padding
                - self.settings.arc_margin
                - arc_radius
            )
        )
        repeat_end_connection_pos = self.width - (
            self._end_arc_size
            if self._need_shift_end_arc
            else (
                self._end_arc_size
                + self._additional_end_padding
                - self.settings.arc_margin
                - arc_radius
            )
        )

        x_pos = (
            self._center_offset
            + self._repeat.start_padding
            + self._repeat_content_width_r
        )
        y_pos = self._item.down + self._vertical_choice_separation + self._repeat.up
        return merge_ridge_lines(
            reverse_ridge_line(self._repeat.bottom_ridge_line, x_pos)
            + Vec(0, y_pos + self._repeat.height),
            RidgeLine(
                -self.height,
                [
                    Vec(repeat_start_connection_pos, self.down),
                    Vec(repeat_end_connection_pos, 0),
                ],
            ),
            cmp=min,
        )

    def __str__(self):
        if isinstance(self._repeat, Skip):
            return (
                f"{self._item}+"
                if self._item.precedence >= self.precedence
                else f"({self._item})+"
            )
        else:
            item = (
                f"{self._item}"
                if self._item.precedence >= self.precedence
                else f"({self._item})"
            )
            repeat = (
                f"{self._repeat}"
                if self._repeat.precedence >= self.precedence
                else f"({self._repeat})"
            )
            return f"{item} ({repeat} {item})*"



from __future__ import annotations

import typing as _t

from syntax_diagrams._impl.render import (
    LayoutContext,
    LayoutSettings,
    Render,
    RenderContext,
)
from syntax_diagrams._impl.tree import Element
from syntax_diagrams._impl.vec import Vec

T = _t.TypeVar("T")


class End(Element[T], _t.Generic[T]):
    def __init__(self, reverse: bool = False) -> None:
        self._reverse = reverse

    def _calculate_content_layout(
        self, settings: LayoutSettings[T], context: LayoutContext
    ):
        self._isolate()

        self.display_width = self.content_width = settings.marker_width
        self.up = settings.marker_projected_height
        self.down = settings.marker_projected_height
        self.start_margin = 0  # TODO: use node margin?
        self.end_margin = 0

    def _render_content(self, render: Render[T], context: RenderContext):
        if context.reverse:
            pos = context.pos - Vec(self.content_width, 0)
        else:
            pos = context.pos
        if self._reverse ^ context.reverse:
            render.right_marker(pos)
        else:
            render.left_marker(pos)

    def __str__(self) -> str:
        return "^" if self._reverse else "$"



from __future__ import annotations

import math
import typing as _t
from contextlib import contextmanager
from dataclasses import replace
from functools import cached_property

from syntax_diagrams._impl.render import (
    ConnectionDirection,
    ConnectionType,
    LayoutContext,
    LayoutSettings,
    Render,
    RenderContext,
)
from syntax_diagrams._impl.ridge_line import RidgeLine
from syntax_diagrams._impl.vec import Vec

T = _t.TypeVar("T")


class Element(_t.Generic[T]):
    """
    Element dimensions
    ------------------

    Elements get information about how they're connected to other elements
    via layout/rendering context. `context.start_connection` indicates how element's
    input is connected, and `context.end_connection` indicates how element's
    output is connected.

    Element can choose to draw connections itself by calling `_isolate` from
    its layout routine, or it can choose to pass this info to its children.

    In either case, element width includes space required to make these connections.
    Space for input connection is called `_start_padding`, and space for output
    connection is called `_end_padding`. Space for element's content is called
    `_content_width`. Together, these three values constitute element's full width,
    available via the `_width` property.

    Generally speaking, element's content width will include space between leftmost
    and rightmost node, while paddings account for all space necessary
    to connect them.

    Additionally, elements have margins. They control minimal spacing between nodes.
    Margin is calculated as a minimal amount of space that should separate element's
    content from other elements' contents.


    Example: normal start connection point

            ┌─────────┐      ┈┈┈
            │         │      │up
         ┈┈>X───────╮ │      ┈┈┈
            │ ╭─────╯ │      │height
            │ ╰───────┼>┈┈   ┈┈┈
            │         │      │down
            └─────────┘      ┈┈┈
        ►┆┈┈┆◄┈┈┈┈┈┈┈►┆┈┈┆◄
         ┆sm┆  width  ┆em┆

    Here, `X` is element's position, as passed to the `_render` method. `sm` and `em`
    are `start_margin` and `end_margin`. `width` is equal to `content_width`.
    `start_padding` and `end_padding` are both zero.


    Example: split start connection point

    .. code-block::

            ┆   width   ┆
            ┆◄┈┈┈┈┈┈┈┈┈►┆
            ┌─┬─────────┐      ┈┈┈
            │ │         │      │up
            X╭Z───────╮ │      ┈┈┈
            │││ ╭─────╯ │      │height
            │││ ╰───────┼>┈┈   ┈┈┈
         ┈┈>Y╯│         │      │down
            └─┴─────────┘      ┈┈┈
          ►┆┈┈┆◄┈┈┈┈┈┈┈►┆┈┈┆◄
           ┆sm┆ c.width ┆em┆

    Here, `X` is element's position, and `Y` is position of the start connection point.
    `sm` and `em` are `start_margin` and `end_margin`. `c.width` is `content_width`.
    `start_padding` is distance between `X` and `Z`.

    Note that element's width includes width of element's content, plus width
    that's required to build line from connection point (Y) to the point where
    element's content actually starts (Z).

    Note also that margin starts at element's content box, not its padding box.


    Example: stack start connection point

    .. code-block::

            ┆   width   ┆
            ┆◄┈┈┈┈┈┈┈┈┈►┆
            ┆╭Y         ┆
            ┆│          ┆
            ┌┼┬─────────┐      ┈┈┈
            │││         │      │up
            X╰Z───────╮ │      ┈┈┈
            │ │ ╭─────╯ │      │height
            │ │ ╰───────┼>┈┈   ┈┈┈
            │ │         │      │down
            └─┴─────────┘      ┈┈┈
          ►┆┈┈┆◄┈┈┈┈┈┈┈►┆┈┈┆◄
           ┆sm┆ c.width ┆em┆

    Here, the situation is similar to the split connection point, but our element
    is a part of a stack. The connection line is coming from above the element.


    Optional enter and exit lines
    -----------------------------

    When an element is part of a larger optional element, it can use optional's
    bypass line to optimize its rendering.

    Consider a diagram for `(A? B)? C`. We can naively render it as such:

    .. code-block::

         ╭────────────╮
         │╭─────╮     │
         ↑↑     ↓     ↓
        ─┴┴─ A ─┴─ B ─┴─ C ──

    However, there's a potential to make it nicer. When element `(A? B)?`
    (a.k.a. outer optional) plans its layout, it communicates to its children
    that there is a bypass line above them. Inner optional can then reuse this line
    to add optional enters and exits, leading to the following result:

    .. code-block::

         ╭─────┬─────╮
         ↑     ↓     ↓
        ─┴─ A ─┴─ B ─┴─ C ──
               ^
               note: connection from optional line
                     into the element

    Specifically, `context.opt_enter_top` and `context.opt_enter_bottom` indicate
    lines that can split off *into the element*, while `context.opt_exit_top`
    and `context.opt_exit_bottom` indicate lines that the element *can split off to*.

    In the above example, the line around `(A? B)?` was `opt_enter_top` line.
    In diagram for `A (B C?)?`, the line around `(B C?)?` is `opt_exit_top`:

    .. code-block::

              ╭─────┬─────╮
              ↑     ↑     ↓
        ── A ─┴─ B ─┴─ C ─┴─
                    ^ note: connection from element
                            into the optional line


    Rendering elements right-to-left
    --------------------------------

    When rendering elements right-to-left (i.e. ``context.reverse=True``),
    everything is simply mirrored. We avoid headaches by implementing code
    for left-to-right rendering, then multiplying all offsets by -1 when rendering
    in reverse. `context.dir` is helpful here.
    """

    display_width: int = 0
    """
    Total width this element takes up on the screen. Can be larger than `_width`
    in case of stack optimizations.

    """

    content_width: int = 0
    """
    Width of element's contents.

    """

    start_padding: int = 0
    """
    Space that's required to connect element's input to its content.

    """

    end_padding: int = 0
    """
    Space that's required to connect element's content to its output.

    """

    start_margin: int = 0
    """
    Minimal space between this node's content box and content boxes of its neighbors.

    """

    end_margin: int = 0
    """
    Minimal space between this node's content box and content boxes of its neighbors.

    """

    height: int = 0
    """
    Difference between level of element's input line and its output line.

    """

    up: int = 0
    """
    Space the element occupies above its input line.

    """

    down: int = 0
    """
    Space the element occupies below its output line.

    """

    settings: LayoutSettings[T] | None = None
    """
    Used to check if settings changed from the last layout calculation,
    and re-calculate layout only if needed.

    This is necessary because auto-wrapping frequently adjusts connection types
    of sequence's children and re-calculates their layout.

    """

    context: LayoutContext | None = None
    """
    Used to check if settings changed from the last layout calculation,
    and re-calculate layout only if needed.

    This is necessary because auto-wrapping frequently adjusts connection types
    of sequence's children and re-calculates their layout.

    """

    @property
    def width(self) -> int:
        """
        Total width of an element.

        """

        return self.start_padding + self.content_width + self.end_padding

    @cached_property
    def precedence(self) -> int:
        """
        Element's precedence, used in `__str__` method.

        """

        return 3

    @cached_property
    def contains_choices(self) -> bool:
        """
        True if there are non-trivial `_Choice` elements within this element.

        Used to choose vertical spacing in certain situations
        (see `vertical_choice_separation_outer` and `vertical_choice_separation`.)

        """

        return False

    @cached_property
    def can_use_opt_enters(self) -> bool:
        """
        Indicates that this element can benefit from giving it an optional enter line.

        """

        return False

    @cached_property
    def can_use_opt_exits(self) -> bool:
        """
        Indicates that this element can benefit from giving it an optional exit line.

        """

        return False

    def _isolate(self, start: bool = True, end: bool = True):
        """
        Calling this method from `_calculate_content_layout` enables automatic handling
        of paddings and element connections.

        If called, `_isolate` returns a new version of context with adjusted enter
        and exit types and available width. After `_calculate_content_layout` finishes,
        element's paddings are expanded to accommodate connections. `_render_content`
        will receive a context with adjusted positions as well.

        """

        self.__isolated_start = start
        self.__isolated_end = end
        assert self.context
        return replace(
            self.context,
            width=max(
                0,
                self.context.width
                - (self.__start_arc_size if start else 0)
                - (self.__end_arc_size if end else 0),
            ),
            start_connection=(
                (
                    ConnectionType.NORMAL
                    if self.__start_connection is not ConnectionType.NULL
                    else ConnectionType.NULL
                )
                if start
                else self.context.start_connection
            ),
            start_top_is_clear=(True if start else self.context.start_top_is_clear),
            start_bottom_is_clear=(
                True if start else self.context.start_bottom_is_clear
            ),
            start_direction=(
                ConnectionDirection.STRAIGHT if start else self.context.start_direction
            ),
            end_connection=(
                (
                    ConnectionType.NORMAL
                    if self.__end_connection is not ConnectionType.NULL
                    else ConnectionType.NULL
                )
                if end
                else self.context.end_connection
            ),
            end_top_is_clear=(True if end else self.context.end_top_is_clear),
            end_bottom_is_clear=(True if end else self.context.end_bottom_is_clear),
            end_direction=(
                ConnectionDirection.STRAIGHT if end else self.context.end_direction
            ),
        )

    def calculate_layout(self, settings: LayoutSettings[T], context: LayoutContext):
        """
        Run layout calculation.

        """

        if settings == self.settings and context == self.context:
            return

        self.settings = settings
        self.context = context
        self.display_width = 0
        self.content_width = 0
        self.start_padding = 0
        self.start_margin = 0
        self.end_margin = 0
        self.end_padding = 0
        self.height = 0
        self.up = 0
        self.down = 0
        self.__dict__.pop("top_ridge_line", None)
        self.__dict__.pop("bottom_ridge_line", None)

        self.__start_connection = context.start_connection
        self.__start_arc_size = self.__start_connection.arc_size(settings)
        self.__end_connection = context.end_connection
        self.__end_arc_size = self.__end_connection.arc_size(settings)

        self.__isolated_start = False
        self.__isolated_end = False
        self._calculate_content_layout(settings, context)

        self.__width_before_adjustments = self.width
        self.__start_padding_before_adjustments = self.start_padding
        self.__end_padding_before_adjustments = self.end_padding
        self.__start_margin_before_adjustments = self.start_margin
        self.__end_margin_before_adjustments = self.end_margin

        if self.__isolated_start:
            self.start_padding += self.__start_arc_size
            self.display_width += self.__start_arc_size
            if self.__start_connection in (ConnectionType.STACK, ConnectionType.SPLIT):
                self.start_margin = max(
                    self.start_padding + settings.arc_margin, self.start_margin
                )
        if self.__isolated_end:
            self.end_padding += self.__end_arc_size
            self.display_width += self.__end_arc_size
            if self.__end_connection in (ConnectionType.STACK, ConnectionType.SPLIT):
                self.end_margin = max(
                    self.end_padding + settings.arc_margin, self.end_margin
                )

    def _calculate_content_layout(
        self, settings: LayoutSettings[T], context: LayoutContext
    ):
        """
        Actual layout logic implementation, called from `_calculate_layout`.

        """

        raise NotImplementedError()

    def render(self, render: Render[T], context: RenderContext):
        """
        Render the element.

        """

        render.enter(self)
        render.debug_ridge_line(context.pos, self, context.reverse)

        if self.__isolated_start:
            start_content_pos = context.pos + Vec(
                context.dir * self.__start_arc_size, 0
            )
            start_connection_pos = start_content_pos
            match self.__start_connection:
                case ConnectionType.NULL:
                    pass
                case ConnectionType.NORMAL | ConnectionType.STACK_BOUND:
                    (
                        (render)
                        .line(
                            context.start_connection_pos,
                            context.reverse,
                            "dbg-isolated-line",
                        )
                        .segment_abs(start_content_pos.x)
                    )
                case ConnectionType.STACK:
                    (
                        (render)
                        .line(
                            context.start_connection_pos,
                            context.reverse,
                            "dbg-isolated-line",
                        )
                        .bend_backward_abs(start_content_pos.y, arrow_begin=True)
                        .segment_abs(start_content_pos.x, arrow_begin=True)
                    )
                case ConnectionType.SPLIT:
                    (
                        (render)
                        .line(
                            context.start_connection_pos,
                            context.reverse,
                            "dbg-isolated-line",
                        )
                        .bend_forward_abs(start_content_pos.y, arrow_begin=True)
                        .segment_abs(start_content_pos.x, arrow_begin=True)
                    )
        else:
            start_content_pos = context.pos
            start_connection_pos = context.start_connection_pos

        if self.__isolated_end:
            end_connection_pos = start_content_pos + Vec(
                context.dir * self.__width_before_adjustments, self.height
            )
        else:
            end_connection_pos = context.end_connection_pos

        content_context = replace(
            context,
            pos=start_content_pos,
            start_connection_pos=start_connection_pos,
            end_connection_pos=end_connection_pos,
        )

        with self.__isolated_context():
            self._render_content(render, content_context)

        if self.__isolated_end:
            match self.__end_connection:
                case ConnectionType.NULL:
                    pass
                case ConnectionType.NORMAL | ConnectionType.STACK_BOUND:
                    (
                        (render)
                        .line(end_connection_pos, context.reverse, "dbg-isolated-line")
                        .segment_abs(context.end_connection_pos.x)
                    )
                case ConnectionType.STACK:
                    (
                        (render)
                        .line(end_connection_pos, context.reverse, "dbg-isolated-line")
                        .segment_abs(context.end_connection_pos.x, arrow_end=True)
                        .bend_backward_reverse_abs(
                            context.end_connection_pos.y, arrow_end=True
                        )
                    )
                case ConnectionType.SPLIT:
                    arc_radius = context.dir * math.ceil(2 * render.settings.arc_radius)
                    (
                        (render)
                        .line(end_connection_pos, context.reverse, "dbg-isolated-line")
                        .segment_abs(
                            context.end_connection_pos.x - arc_radius, arrow_end=True
                        )
                        .bend_forward_abs(context.end_connection_pos.y, arrow_end=True)
                    )

        render.debug(self, context)
        render.debug_pos(context.pos, "dbg-primary-pos")
        render.debug_pos(start_connection_pos, "dbg-isolated-pos")
        render.debug_pos(end_connection_pos, "dbg-isolated-pos")
        for opt in [
            context.opt_enter_top,
            context.opt_enter_bottom,
            context.opt_exit_top,
            context.opt_exit_bottom,
        ]:
            if opt:
                for pos in opt[1:]:
                    if pos:
                        render.debug_pos(pos, "dbg-alternative-pos")
        render.debug_pos(context.start_connection_pos, "dbg-primary-pos")
        render.debug_pos(context.end_connection_pos, "dbg-primary-pos")
        render.exit()

    def _render_content(self, render: Render[T], context: RenderContext):
        """
        Actual element rendering, called from `_render`.

        """

        raise NotImplementedError()

    @cached_property
    def top_ridge_line(self) -> RidgeLine:
        """
        Distances from input line to topmost element at the given coordinate.

        """

        with self.__isolated_context():
            ridge_line = self._calculate_top_ridge_line()
        if self.__isolated_start and self.__start_arc_size > 0:
            ridge_line = ridge_line + Vec(self.__start_arc_size, 0)
        return ridge_line

    def _calculate_top_ridge_line(self) -> RidgeLine:
        return RidgeLine(
            0,
            [
                Vec(self.start_padding, self.up),
                Vec(self.width - self.end_padding, -self.height),
            ],
        )

    @cached_property
    def bottom_ridge_line(self) -> RidgeLine:
        """
        Distances from output line to bottommost element at the given coordinate.

        """

        with self.__isolated_context():
            ridge_line = self._calculate_bottom_ridge_line()
        if self.__isolated_start and self.__start_arc_size > 0:
            ridge_line = ridge_line + Vec(self.__start_arc_size, 0)
        return ridge_line

    def _calculate_bottom_ridge_line(self) -> RidgeLine:
        return RidgeLine(
            -self.height,
            [
                Vec(self.start_padding, self.down),
                Vec(self.width - self.end_padding, 0),
            ],
        )

    @contextmanager
    def __isolated_context(self):
        start_padding = self.start_padding
        self.start_padding = self.__start_padding_before_adjustments
        end_padding = self.end_padding
        self.end_padding = self.__end_padding_before_adjustments
        start_margin = self.start_margin
        self.start_margin = self.__start_margin_before_adjustments
        end_margin = self.end_margin
        self.end_margin = self.__end_margin_before_adjustments

        yield

        self.start_padding = start_padding
        self.end_padding = end_padding
        self.start_margin = start_margin
        self.end_margin = end_margin



import sys

from sybil import Sybil
from sybil.parsers.codeblock import PythonCodeBlockParser
from sybil.parsers.doctest import DocTestParser
from sybil.parsers.rest import SkipParser

if sys.version_info >= (3, 10):
    # Examples use the new typing syntax.
    pytest_collect_file = Sybil(
        parsers=[
            DocTestParser(),
            PythonCodeBlockParser(),
            SkipParser(),
        ],
        patterns=["*.rst", "*.py"],
    ).pytest()



import datetime

import syntax_diagrams

project = "Syntax Diagrams"
copyright = f"{datetime.date.today().year}, Tamika Nomara"
author = "Tamika Nomara"
release = version = syntax_diagrams.__version__

del syntax_diagrams

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.githubpages",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autodoc",
    "sphinx_design",
    "sphinx_syntax",
]

templates_path = ["_templates"]
exclude_patterns = []

primary_domain = "py"
default_role = "py:obj"
autodoc_type_aliases = {"Element": "syntax_diagrams.Element"}
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "PIL": ("https://pillow.readthedocs.io/en/stable/", None),
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_extra_path = ["_extra/robots.txt"]
html_css_files = ["syntax-diagrams-ext.css"]
html_theme_options = {
    "source_repository": "https://github.com/taminomara/syntax-diagrams",
    "source_branch": "main",
    "source_directory": "docs/source",
}



"""
This type stub file was generated by pyright.
"""

from typing import *

__all__ = [
    "wcwidth",
    "wcswidth",
]

def wcwidth(wc: str, unicode_version: str = "auto") -> int:
    r"""
    Given one Unicode character, return its printable length on a terminal.

    :param str wc: A single Unicode character.
    :param str unicode_version: A Unicode version number, such as
        ``'6.0.0'``. A list of version levels suported by wcwidth
        is returned by :func:`list_versions`.

        Any version string may be specified without error -- the nearest
        matching version is selected.  When ``latest`` (default), the
        highest Unicode version level is used.
    :return: The width, in cells, necessary to display the character of
        Unicode string character, ``wc``.  Returns 0 if the ``wc`` argument has
        no printable effect on a terminal (such as NUL '\0'), -1 if ``wc`` is
        not printable, or has an indeterminate effect on the terminal, such as
        a control character.  Otherwise, the number of column positions the
        character occupies on a graphic terminal (1 or 2) is returned.
    :rtype: int

    See :ref:`Specification` for details of cell measurement.

    """
    ...

def wcswidth(pwcs: str, n: int = ..., unicode_version: str = "auto") -> int:
    """
    Given a unicode string, return its printable length on a terminal.

    :param str pwcs: Measure width of given unicode string.
    :param int n: When ``n`` is None (default), return the length of the entire
        string, otherwise only the first ``n`` characters are measured. This
        argument exists only for compatibility with the C POSIX function
        signature. It is suggested instead to use python's string slicing
        capability, ``wcswidth(pwcs[:n])``
    :param str unicode_version: An explicit definition of the unicode version
        level to use for determination, may be ``auto`` (default), which uses
        the Environment Variable, ``UNICODE_VERSION`` if defined, or the latest
        available unicode version, otherwise.
    :rtype: int
    :returns: The width, in cells, needed to display the first ``n`` characters
        of the unicode string ``pwcs``.  Returns ``-1`` for C0 and C1 control
        characters!

    See :ref:`Specification` for details of cell measurement.

    """
    ...



from __future__ import annotations

import dataclasses
import json
import math
import sys
import typing as _t
from dataclasses import dataclass
from enum import Enum
from functools import cached_property

from syntax_diagrams._impl.vec import Vec
from syntax_diagrams.measure import TextMeasure
from syntax_diagrams.render import EndClass
from syntax_diagrams.resolver import HrefResolver

if _t.TYPE_CHECKING:
    from syntax_diagrams._impl.tree import Element

T = _t.TypeVar("T")


_DEBUG_ATTRS = [
    "display_width",
    "width",
    "content_width",
    "start_padding",
    "end_padding",
    "start_margin",
    "end_margin",
    "height",
    "up",
    "down",
    "context",
]

_IGNORE_ATTRS = {"settings"}


class Render(_t.Generic[T]):
    def __init__(self, settings: LayoutSettings[T], dump_debug_data: bool):
        self.settings = settings
        self._debug = dump_debug_data
        self._id = 0
        self._ids: dict[Element[T], str] = {}
        self._debug_data: dict[str, _t.Any] = {}
        self._debug_stack: list[str] = []

    def _make_debug_id(self, elem: Element[T]) -> str:
        if elem not in self._ids:
            self._ids[elem] = str(self._id)
            self._id += 1
        return self._ids[elem]

    def _debug_parent(self) -> str | None:
        return self._debug_stack[-1] if self._debug_stack else None

    def write(self, f=sys.stdout):
        raise NotImplementedError()

    def to_string(self):
        raise NotImplementedError()

    def debug_data(self) -> str:
        return self._DebugJSONEncoder(self).encode(
            dict(rendered=self.to_string(), debug_data=self._debug_data)
        )

    def enter(self, node: Element[_t.Any]):
        if self._debug:
            node_id = self._make_debug_id(node)
            name = node.__class__.__name__.lower()
            if hasattr(node, "_text"):
                name += f" {json.dumps(getattr(node, "_text"))}"
            data = {}
            for k in _DEBUG_ATTRS:
                data[k] = getattr(node, k, None)
            for k, v in vars(node).items():
                if k.startswith("_Element__"):
                    k = "__" + k[len("_Element__") :]
                if k not in _IGNORE_ATTRS and k not in data:
                    data[k] = v
            data["$order"] = list(data.keys())
            self._debug_data[node_id] = {
                "parent": self._debug_parent(),
                "index": self._id,
                "name": name,
                "data": data,
            }
            self._debug_stack.append(node_id)

    def exit(self):
        if self._debug:
            self._debug_stack.pop()

    def line(self, pos: Vec, reverse: bool = False, css_class: str = "") -> Line:
        raise NotImplementedError()

    def node(
        self,
        pos: Vec,
        style: NodeStyle,
        css_class: str | None,
        content_width: int,
        up: int,
        down: int,
        radius: int,
        padding: int,
        text_width: int,
        text_height: int,
        text: str,
        href: str | None,
        title: str | None,
    ):
        raise NotImplementedError()

    def group(
        self,
        pos: Vec,
        width: int,
        height: int,
        css_class: str | None,
        text_width: int,
        text_height: int,
        text: str | None,
        href: str | None,
        title: str | None,
    ):
        raise NotImplementedError()

    def left_marker(self, pos: Vec):
        raise NotImplementedError()

    def right_marker(self, pos: Vec):
        raise NotImplementedError()

    def debug(self, node: Element[_t.Any], context: RenderContext):
        pass

    def debug_pos(self, pos: Vec, css_class: str = ""):
        pass

    def debug_ridge_line(self, pos: Vec, node: Element[_t.Any], reverse: bool):
        pass

    class _DebugJSONEncoder(json.JSONEncoder):
        def __init__(self, render: Render[_t.Any]):
            super().__init__()
            self._render = render

        def default(self, o):
            from syntax_diagrams._impl.tree import Element

            if dataclasses.is_dataclass(o) and not isinstance(o, type):
                d = dataclasses.asdict(o)
                d["$order"] = list(d.keys())
                return d
            elif isinstance(o, Enum):
                return o.value
            elif isinstance(o, Element):
                return {
                    "$elem": _t.cast(Element[_t.Any], o).__class__.__qualname__,
                    "$id": self._render._make_debug_id(o),
                }
            return super().default(o)


Direction: _t.TypeAlias = _t.Literal["w", "e"]


class Line:
    _reverse: bool

    def segment_abs(
        self, x: int, arrow_begin: bool = False, arrow_end: bool = False
    ) -> Line:
        raise NotImplementedError()

    def bend_forward_abs(
        self,
        y: int,
        arrow_begin: bool = False,
        arrow_end: bool = False,
    ) -> Line:
        d = "e" if not self._reverse else "w"
        return self.bend(y, self.reverse_direction(d), d, arrow_begin, arrow_end)

    def bend_backward_abs(
        self,
        y: int,
        arrow_begin: bool = False,
        arrow_end: bool = False,
    ) -> Line:
        d = "e" if not self._reverse else "w"
        return self.bend(y, d, d, arrow_begin, arrow_end)

    def bend_backward_reverse_abs(
        self,
        y: int,
        arrow_begin: bool = False,
        arrow_end: bool = False,
    ) -> Line:
        d = "w" if not self._reverse else "e"
        return self.bend(y, d, d, arrow_begin, arrow_end)

    def bend(
        self,
        y: int,
        coming_from: Direction,
        coming_to: Direction | None,
        arrow_begin: bool = False,
        arrow_end: bool = False,
    ) -> Line:
        raise NotADirectoryError()

    @staticmethod
    def reverse_direction(d: Direction):
        return "e" if d == "w" else "w"


@dataclass
class LayoutSettings(_t.Generic[T]):
    horizontal_seq_separation: int
    vertical_choice_separation_outer: int
    vertical_choice_separation: int
    vertical_seq_separation_outer: int
    vertical_seq_separation: int

    arc_radius: int | float
    arc_margin: int

    terminal_text_measure: TextMeasure
    terminal_horizontal_padding: int
    terminal_vertical_padding: int
    terminal_radius: int

    non_terminal_text_measure: TextMeasure
    non_terminal_horizontal_padding: int
    non_terminal_vertical_padding: int
    non_terminal_radius: int

    comment_text_measure: TextMeasure
    comment_horizontal_padding: int
    comment_vertical_padding: int
    comment_radius: int

    group_text_measure: TextMeasure
    group_vertical_padding: int
    group_horizontal_padding: int
    group_vertical_margin: int
    group_horizontal_margin: int
    group_thickness: int
    group_radius: int
    group_text_vertical_offset: int
    group_text_horizontal_offset: int

    marker_width: int
    marker_projected_height: int

    end_class: EndClass

    hidden_symbol_escape: tuple[str, str]

    href_resolver: HrefResolver[T] = HrefResolver()


class NodeStyle(Enum):
    """
    Style used to draw a node.
    """

    TERMINAL = "TERMINAL"
    NON_TERMINAL = "NON_TERMINAL"
    COMMENT = "COMMENT"


class ConnectionType(Enum):
    """
    Kind of curve that connects this element to the next/previous one.

    """

    NULL = "NULL"
    """
    Element is not connected. Happens at the edges of diagram.

    .. code-block::

       NODE

            ^ end connection is NULL

    """

    NORMAL = "NORMAL"
    """
    Element is connected by a straight line.

    .. code-block::

       NODE ──

            ^ end connection is NORMAL

    """

    STACK_BOUND = "STACK_BOUND"
    """
    Element is connected by a straight line. Like `NORMAL`, but longer.

    .. code-block::

       NODE ────

            ^ end connection is STACK_BOUND

    """

    STACK = "STACK"
    """
    Element is attached to a stack, i.e. the line makes a reverse turn before/after it.

    .. code-block::

            ╮
            ↑
       NODE ╯

            ^ end connection is STACK

    """

    SPLIT = "SPLIT"
    """
    Element is attached to a choice, i.e. the line makes a forward turn before/after it.

    .. code-block::

            ╭
            ↑
       NODE ╯

            ^ end connection is SPLIT

    """

    def arc_size(self, settings: LayoutSettings[_t.Any]) -> int:
        match self:
            case ConnectionType.NORMAL | ConnectionType.NULL:
                return 0
            case ConnectionType.STACK | ConnectionType.STACK_BOUND:
                return math.ceil(settings.arc_radius) + settings.arc_margin
            case ConnectionType.SPLIT:
                return math.ceil(2 * settings.arc_radius) + settings.arc_margin


class ConnectionDirection(Enum):
    """
    Indicates where connection makes a turn before/after the element.

    """

    UP = "UP"
    """
    Connection goes up.

    """

    DOWN = "DOWN"
    """
    Connection goes down.

    """

    STRAIGHT = "STRAIGHT"
    """
    Connection goes straight.

    """


ClearOfIntersections: _t.TypeAlias = bool
"""
Indicates whether element's connection intersects with other connections.

.. code-block::

         ╮
         ↑
    NODE ╯

        ^ end connection is STACK and it's clear of any intersections.

.. code-block::

         ╮
         ↑ <┈ end connection goes up
    NODE ┤
         ↑ <┈ there are some intersections, meaning that
              OneOrMore can't reuse end connection for its repeat loop.

         ^ end connection is STACK and there are intersections.

"""


@dataclass(kw_only=True)
class LayoutContext:
    width: int
    is_outer: bool
    start_connection: ConnectionType = ConnectionType.NORMAL
    start_top_is_clear: ClearOfIntersections = False
    start_bottom_is_clear: ClearOfIntersections = False
    start_direction: ConnectionDirection = ConnectionDirection.STRAIGHT
    end_connection: ConnectionType = ConnectionType.NORMAL
    end_top_is_clear: ClearOfIntersections = False
    end_bottom_is_clear: ClearOfIntersections = False
    end_direction: ConnectionDirection = ConnectionDirection.STRAIGHT
    allow_shrinking_stacks: bool = False
    opt_enter_top: bool = False
    opt_enter_bottom: bool = False
    opt_exit_top: bool = False
    opt_exit_bottom: bool = False


@dataclass(kw_only=True)
class RenderContext:
    pos: Vec
    start_connection_pos: Vec
    end_connection_pos: Vec
    reverse: bool
    opt_enter_top: tuple[Direction, Vec] | None = None
    opt_enter_bottom: tuple[Direction, Vec] | None = None
    opt_exit_top: tuple[Direction, Vec, Vec | None] | None = None
    opt_exit_bottom: tuple[Direction, Vec, Vec | None] | None = None

    @cached_property
    def dir(self):
        # Direction
        return 1 if not self.reverse else -1



from __future__ import annotations

import math
import typing as _t
from functools import cached_property

from syntax_diagrams._impl.render import (
    ConnectionDirection,
    ConnectionType,
    LayoutContext,
    LayoutSettings,
    Render,
    RenderContext,
)
from syntax_diagrams._impl.ridge_line import (
    RidgeLine,
    find_distance,
    merge_ridge_lines,
)
from syntax_diagrams._impl.tree import Element
from syntax_diagrams._impl.tree.skip import Skip
from syntax_diagrams._impl.vec import Vec

T = _t.TypeVar("T")


class Choice(Element[T], _t.Generic[T]):
    _default: int
    _items: list[Element[T]]
    _has_skip: bool
    _is_optional: bool

    _connect_opt_enter: bool
    _connect_opt_exit: bool
    _layout: list[tuple[Element[T], int]]
    _start_connection: ConnectionType
    _end_connection: ConnectionType
    _upper_rail_can_use_added_opt_enters: bool
    _upper_rail_can_use_added_opt_exits: bool
    _lower_rail_can_use_added_opt_enters: bool
    _lower_rail_can_use_added_opt_exits: bool

    def __new__(cls, items: list[Element[T]], default: int) -> Element[T]:
        if len(items) == 0:
            return Skip()
        elif len(items) == 1:
            return items[0]

        assert 0 <= default < len(items)

        merged_default = 0
        merged_items: list[Element[T]] = []
        for i, item in enumerate(items):
            if isinstance(item, Choice):
                if i == default:
                    merged_default = len(merged_items) + item._default
                merged_items.extend(item._items)
            else:
                if i == default:
                    merged_default = len(merged_items)
                merged_items.append(item)

        filtered_items: list[Element[T]] = []
        filtered_default = 0
        seen_skip = isinstance(merged_items[merged_default], Skip)
        for i, item in enumerate(merged_items):
            if i == merged_default:
                filtered_default = len(filtered_items)
            elif isinstance(item, Skip):
                if seen_skip:
                    continue
                else:
                    seen_skip = True
            filtered_items.append(item)

        if len(filtered_items) == 1:
            return filtered_items[0]

        self = super().__new__(cls)
        self._items = filtered_items
        self._default = filtered_default
        self._has_skip = seen_skip
        self._is_optional = seen_skip and len(self._items) == 2

        return self

    @cached_property
    def precedence(self) -> int:
        # Optional renders as an unary operator
        return 3 if self._has_skip else 1

    @cached_property
    def contains_choices(self) -> bool:
        # Optionals are not true choices
        return not self._is_optional

    @cached_property
    def can_use_opt_enters(self) -> bool:
        return self._has_skip

    @cached_property
    def can_use_opt_exits(self) -> bool:
        return self._has_skip

    def _calculate_content_layout(
        self, settings: LayoutSettings[T], context: LayoutContext
    ):
        skip_skips = False
        self._connect_opt_enter = False
        self._connect_opt_exit = False
        if self._has_skip:
            if (context.opt_enter_top and context.opt_exit_top) or (
                context.opt_enter_bottom and context.opt_exit_bottom
            ):
                # If we have enter and exit on the same side, they belong to the same
                # line, so we can skip skips without adding new connections.
                skip_skips = True
            else:
                # Otherwise, we'll note both. We can choose either of them later.
                if context.opt_enter_top or context.opt_enter_bottom:
                    skip_skips = True
                    self._connect_opt_enter = True
                if context.opt_exit_top or context.opt_exit_bottom:
                    skip_skips = True
                    self._connect_opt_exit = True

        if skip_skips:
            items: list[Element[T]] = []
            default = 0
            for i, item in enumerate(self._items):
                if i == self._default:
                    default = len(items)
                if not isinstance(item, Skip):
                    items.append(item)
            if default >= len(items):
                default = len(items) - 1
        else:
            default = self._default
            items = self._items

        if len(items) == 1:
            self._start_connection = context.start_connection
            if (
                self._start_connection
                in (
                    ConnectionType.NORMAL,
                    ConnectionType.STACK_BOUND,
                )
                and self._connect_opt_exit
            ):
                self._start_connection = ConnectionType.SPLIT
            self._end_connection = context.end_connection
            if (
                self._end_connection
                in (
                    ConnectionType.NORMAL,
                    ConnectionType.STACK_BOUND,
                )
                and self._connect_opt_enter
            ):
                self._end_connection = ConnectionType.SPLIT
        else:
            self._start_connection = context.start_connection
            if self._start_connection in (
                ConnectionType.NORMAL,
                ConnectionType.STACK_BOUND,
            ):
                self._start_connection = ConnectionType.SPLIT
            self._end_connection = context.end_connection
            if self._end_connection in (
                ConnectionType.NORMAL,
                ConnectionType.STACK_BOUND,
            ):
                self._end_connection = ConnectionType.SPLIT

            self._upper_rail_can_use_added_opt_enters = False
            self._upper_rail_can_use_added_opt_exits = False
            self._lower_rail_can_use_added_opt_enters = False
            self._lower_rail_can_use_added_opt_exits = False

            for i, item in enumerate(items):
                if isinstance(item, Skip):
                    if i > 0:
                        upper_rail = items[i - 1]
                        self._upper_rail_can_use_added_opt_enters = (
                            upper_rail.can_use_opt_enters
                        )
                        self._upper_rail_can_use_added_opt_exits = (
                            upper_rail.can_use_opt_exits
                        )
                    if i < len(items) - 1:
                        lower_rail = items[i + 1]
                        self._lower_rail_can_use_added_opt_enters = (
                            lower_rail.can_use_opt_enters
                        )
                        self._lower_rail_can_use_added_opt_exits = (
                            lower_rail.can_use_opt_exits
                        )
                    break

            # Prevent optimizations that can break diagram's correctness: if one rail
            # uses an opt enter, another rail shouldn't use an opt exit
            # to the same opt line (and vice versa). Otherwise we risk drawing
            # `((AA? B) | (C DD?))?` like this:
            #
            #       ╭─ AA ─┬─ B ─╮
            #       ↑      ↑     ↓
            #     ──┼─────┬┴─────┼──
            #       ↓     ↑      ↑
            #       ╰─ C ─┴─ DD ─╯
            #
            # Here, `AA?` and `DD?` reuse the same opt line, leading to input `C B`
            # being erroneously allowed by the diagram.
            if self._upper_rail_can_use_added_opt_enters:
                self._lower_rail_can_use_added_opt_exits = False
            if self._lower_rail_can_use_added_opt_enters:
                self._upper_rail_can_use_added_opt_exits = False
            if self._upper_rail_can_use_added_opt_exits:
                self._lower_rail_can_use_added_opt_enters = False
            if self._lower_rail_can_use_added_opt_exits:
                self._upper_rail_can_use_added_opt_enters = False

        if any(item.contains_choices for item in self._items):
            vertical_separation = settings.vertical_choice_separation_outer
        else:
            vertical_separation = settings.vertical_choice_separation

        # Total element's width.
        width = 0
        display_width = 0
        # Padding between stack's left border and nearest content box.
        start_padding = None
        # Leftmost margin relative to the stack's left border. Can be negative.
        start_margin_offset = None
        # Padding between stack's right border and nearest content box.
        end_padding = None
        # Rightmost margin relative to the stack's left border.
        end_margin_offset = None

        current_pos = 0

        self._layout = []

        for i, item in enumerate(items):
            if i < default:
                direction = ConnectionDirection.DOWN
            elif i > default:
                direction = ConnectionDirection.UP
            else:
                direction = ConnectionDirection.STRAIGHT

            line_context = LayoutContext(
                width=context.width,
                is_outer=False,
                start_connection=self._start_connection,
                start_top_is_clear=(
                    context.start_top_is_clear
                    if (
                        i == 0
                        and (not self._connect_opt_exit or not context.opt_exit_top)
                    )
                    else False
                ),
                start_bottom_is_clear=(
                    context.start_bottom_is_clear
                    if (
                        i == len(items) - 1
                        and (not self._connect_opt_exit or not context.opt_exit_bottom)
                    )
                    else False
                ),
                start_direction=direction,
                end_connection=self._end_connection,
                end_top_is_clear=(
                    context.end_top_is_clear
                    if (
                        i == 0
                        and (not self._connect_opt_enter or not context.opt_enter_top)
                    )
                    else False
                ),
                end_bottom_is_clear=(
                    context.end_bottom_is_clear
                    if (
                        i == len(items) - 1
                        and (
                            not self._connect_opt_enter or not context.opt_enter_bottom
                        )
                    )
                    else False
                ),
                end_direction=direction,
                allow_shrinking_stacks=context.allow_shrinking_stacks and i == 0,
            )

            if i == 0:
                line_context.opt_enter_top = context.opt_enter_top
                line_context.opt_exit_top = context.opt_exit_top
            elif isinstance(items[i - 1], Skip):
                if self._lower_rail_can_use_added_opt_enters:
                    line_context.opt_enter_top = True
                if self._lower_rail_can_use_added_opt_exits:
                    line_context.opt_exit_top = True
            if i == len(items) - 1:
                line_context.opt_enter_bottom = context.opt_enter_bottom
                line_context.opt_exit_bottom = context.opt_exit_bottom
            elif isinstance(items[i + 1], Skip):
                if self._upper_rail_can_use_added_opt_enters:
                    line_context.opt_enter_bottom = True
                if self._upper_rail_can_use_added_opt_exits:
                    line_context.opt_exit_bottom = True

            item.calculate_layout(settings, line_context)

            if i == 0:
                current_pos += item.up
            else:
                current_pos += (
                    find_distance(items[i - 1].bottom_ridge_line, item.top_ridge_line)
                    + vertical_separation
                )

            if i == default:
                self._layout.append((item, current_pos))

                self.up = current_pos
                self.height = item.height

                for j, (item, pos) in enumerate(self._layout):
                    self._layout[j] = (item, pos - current_pos)

                current_pos = 0
            else:
                self._layout.append((item, current_pos))

            current_pos += item.height

            width = max(width, item.width)
            display_width = max(display_width, item.display_width)

            start_padding = (
                item.start_padding
                if start_padding is None
                else min(start_padding, item.start_padding)
            )
            start_margin_offset = (
                -item.start_margin + item.start_padding
                if start_margin_offset is None
                else min(
                    start_margin_offset,
                    -item.start_margin + item.start_padding,
                )
            )
            end_padding = (
                item.end_padding
                if end_padding is None
                else min(end_padding, item.end_padding)
            )
            end_margin_offset = (
                item.width + item.end_margin - item.end_padding
                if end_margin_offset is None
                else max(
                    end_margin_offset,
                    item.width + item.end_margin - item.end_padding,
                )
            )

        self.down = current_pos + items[-1].down - self.height

        start_padding = start_padding or 0
        end_padding = end_padding or 0

        self.display_width = display_width
        self.content_width = max(0, width - start_padding - end_padding)
        self.start_padding = start_padding
        self.end_padding = end_padding
        if start_margin_offset is None:
            self.start_margin = 0
        else:
            self.start_margin = max(0, (-start_margin_offset) + start_padding)
        if end_margin_offset is None:
            self.end_margin = 0
        else:
            self.end_margin = max(0, (end_margin_offset - width) + end_padding)

        if self._connect_opt_exit:
            self.start_margin = max(
                self.start_margin, settings.arc_margin + self.start_padding
            )
        if self._connect_opt_enter:
            self.end_margin = max(
                self.end_margin, settings.arc_margin + self.end_padding
            )

    def _render_content(self, render: Render[T], context: RenderContext):
        start_arc_size = self._start_connection.arc_size(render.settings)
        end_arc_size = self._end_connection.arc_size(render.settings)

        for i, (item, pos) in enumerate(self._layout):
            line_context = RenderContext(
                pos=context.pos + Vec(0, pos),
                start_connection_pos=context.start_connection_pos,
                end_connection_pos=context.end_connection_pos,
                reverse=context.reverse,
            )

            if i == 0:
                line_context.opt_enter_top = context.opt_enter_top
                line_context.opt_exit_top = context.opt_exit_top
            elif isinstance(self._layout[i - 1][0], Skip):
                line_pos = context.pos.y + self._layout[i - 1][1]
                if self._lower_rail_can_use_added_opt_enters:
                    line_context.opt_enter_top = (
                        "w" if not context.reverse else "e",
                        Vec(context.pos.x + context.dir * start_arc_size, line_pos),
                    )
                if self._lower_rail_can_use_added_opt_exits:
                    line_context.opt_exit_top = (
                        "e" if not context.reverse else "w",
                        Vec(
                            context.pos.x + context.dir * (self.width - end_arc_size),
                            line_pos,
                        ),
                        None,
                    )
            if i == len(self._layout) - 1:
                line_context.opt_enter_bottom = context.opt_enter_bottom
                line_context.opt_exit_bottom = context.opt_exit_bottom
            elif isinstance(self._layout[i + 1][0], Skip):
                line_pos = context.pos.y + self._layout[i + 1][1]
                if self._upper_rail_can_use_added_opt_enters:
                    line_context.opt_enter_bottom = (
                        "w" if not context.reverse else "e",
                        Vec(context.pos.x + context.dir * start_arc_size, line_pos),
                    )
                if self._upper_rail_can_use_added_opt_exits:
                    line_context.opt_exit_bottom = (
                        "e" if not context.reverse else "w",
                        Vec(
                            context.pos.x + context.dir * (self.width - end_arc_size),
                            line_pos,
                        ),
                        None,
                    )

            item.render(render, line_context)

        if self._connect_opt_enter and not (
            self._connect_opt_exit
            and context.opt_exit_bottom
            and context.opt_exit_bottom[2] is not None
        ):
            # Choose exit line.
            opt_enter = context.opt_enter_top or context.opt_enter_bottom
            assert opt_enter
            coming_to, opt_enter_pos = opt_enter

            # Determine directions and where the vertical line will end up.
            vertical_line_x = context.end_connection_pos.x
            if self._end_connection is ConnectionType.STACK:
                coming_from = "w" if not context.reverse else "e"
                if coming_from != coming_to:
                    vertical_line_x += context.dir * math.ceil(
                        2 * render.settings.arc_radius
                    )
            else:
                coming_from = "e" if not context.reverse else "w"
                if coming_from != coming_to:
                    vertical_line_x -= context.dir * math.ceil(
                        2 * render.settings.arc_radius
                    )

            (
                (render)
                .line(opt_enter_pos, context.reverse, "dbg-alternative-pos")
                .segment_abs(
                    vertical_line_x,
                    arrow_begin=True,
                    arrow_end=True,
                )
                .bend(
                    context.end_connection_pos.y,
                    coming_to,
                    coming_from,
                    arrow_begin=True,
                    arrow_end=True,
                )
            )
        elif self._connect_opt_exit:
            # Choose exit line.
            if context.opt_exit_bottom and context.opt_exit_bottom[2] is not None:
                # Prefer bottom if alternative position is available.
                opt_exit = context.opt_exit_bottom
            if context.opt_exit_top:
                # Otherwise, prefer top if it's there.
                opt_exit = context.opt_exit_top
            else:
                # Otherwise, use bottom.
                opt_exit = context.opt_exit_bottom
            assert opt_exit
            coming_to, opt_exit_pos, opt_exit_pos_alt = opt_exit

            # Determine directions and where the vertical line will end up.
            vertical_line_x = context.start_connection_pos.x
            if self._start_connection is ConnectionType.STACK:
                coming_from = "e" if not context.reverse else "w"
                vertical_line_x -= context.dir * render.settings.arc_radius
            else:
                coming_from = "w" if not context.reverse else "e"
                vertical_line_x += context.dir * render.settings.arc_radius

            # Special case for skipping stack rows: position of the vertical line
            # is close to the alternative exit.
            if opt_exit_pos_alt and (
                abs(opt_exit_pos_alt.x - vertical_line_x) <= render.settings.arc_radius
            ):
                opt_exit_pos = opt_exit_pos_alt
                coming_to = None

            # Draw the line.
            (
                (render)
                .line(
                    context.start_connection_pos, context.reverse, "dbg-alternative-pos"
                )
                .bend(
                    opt_exit_pos.y,
                    coming_from,
                    coming_to,
                    arrow_begin=True,
                    arrow_end=True,
                )
                .segment_abs(
                    opt_exit_pos.x,
                    arrow_begin=True,
                    arrow_end=True,
                )
            )

    def _calculate_top_ridge_line(self) -> RidgeLine:
        elem, pos = self._layout[0]
        return merge_ridge_lines(
            merge_ridge_lines(
                elem.top_ridge_line - Vec(0, pos),
                RidgeLine(
                    0,
                    [
                        Vec(0, -pos),
                        Vec(self.start_padding, -pos - elem.height),
                        Vec(self.width, -self.height),
                    ],
                ),
            ),
            RidgeLine(
                0,
                [
                    Vec(0, self.up),
                    Vec(self.display_width, -self.height),
                ],
            ),
            cmp=min,
        )

    def _calculate_bottom_ridge_line(self) -> RidgeLine:
        elem, pos = self._layout[-1]
        pos -= self.height
        return merge_ridge_lines(
            merge_ridge_lines(
                elem.bottom_ridge_line + Vec(0, pos + elem.height),
                RidgeLine(
                    -self.height,
                    [
                        Vec(0, pos),
                        Vec(self.start_padding, pos + elem.height),
                        Vec(self.width, 0),
                    ],
                ),
            ),
            RidgeLine(
                -self.height,
                [
                    Vec(0, self.down),
                    Vec(self.display_width, 0),
                ],
            ),
            cmp=min,
        )

    def __str__(self):
        is_optional = any(isinstance(item, Skip) for item in self._items)
        items = [
            f"{item}" if item.precedence >= self.precedence else f"({item})"
            for item in self._items
            if not isinstance(item, Skip)
        ]
        if is_optional and len(items) == 1:
            if items[0].endswith("+"):
                return items[0][:-1] + "*"
            else:
                return items[0] + "?"
        elif is_optional:
            return "(" + " | ".join(items) + ")?"
        else:
            return " | ".join(items)



from __future__ import annotations

import typing as _t
from dataclasses import replace
from functools import cached_property

from syntax_diagrams._impl.render import (
    LayoutContext,
    LayoutSettings,
    Render,
    RenderContext,
)
from syntax_diagrams._impl.ridge_line import RidgeLine
from syntax_diagrams._impl.tree import Element
from syntax_diagrams._impl.tree.skip import Skip

T = _t.TypeVar("T")


class Barrier(Element[T], _t.Generic[T]):
    _item: Element[T]

    def __new__(cls, item: Element[T]) -> Element[T]:
        if isinstance(item, (Barrier, Skip)):
            return item

        self = super().__new__(cls)
        self._item = item
        return self

    @cached_property
    def precedence(self) -> int:
        return self._item.precedence

    def _calculate_content_layout(
        self, settings: LayoutSettings[T], context: LayoutContext
    ):
        context = self._isolate()
        context = replace(
            context,
            opt_enter_top=None,
            opt_enter_bottom=None,
            opt_exit_top=None,
            opt_exit_bottom=None,
        )

        self._item.calculate_layout(settings, context)

        self.display_width = self._item.display_width
        self.content_width = self._item.content_width
        self.start_padding = self._item.start_padding
        self.end_padding = self._item.end_padding
        self.start_margin = self._item.start_margin
        self.end_margin = self._item.end_margin
        self.height = self._item.height
        self.up = self._item.up
        self.down = self._item.down

    def _render_content(self, render: Render[T], context: RenderContext):
        context = replace(
            context,
            opt_enter_top=None,
            opt_enter_bottom=None,
            opt_exit_top=None,
            opt_exit_bottom=None,
        )

        self._item.render(render, context)

    def _calculate_top_ridge_line(self) -> RidgeLine:
        return self._item.top_ridge_line

    def _calculate_bottom_ridge_line(self) -> RidgeLine:
        return self._item.bottom_ridge_line

    def __str__(self) -> str:
        return str(self._item)



from __future__ import annotations

from syntax_diagrams._version import *
from syntax_diagrams.element import *
from syntax_diagrams.measure import *
from syntax_diagrams.render import *
from syntax_diagrams.resolver import *



import dataclasses

import pytest
import yaml

import syntax_diagrams as rr


def check_svg(
    diagram,
    reverse,
    regression,
    text_regression,
    svg_render_settings,
    text_render_settings,
):
    rendered = rr.render_svg(
        diagram,
        max_width=400,
        reverse=reverse,
        settings=svg_render_settings,
    )
    regression(rendered)


def check_svg_arc_margin(
    diagram,
    reverse,
    regression,
    text_regression,
    svg_render_settings,
    text_render_settings,
):
    svg_render_settings = dataclasses.replace(
        svg_render_settings,
        arc_radius=0,
        arc_margin=50,
        horizontal_seq_separation=0,
    )
    check_svg(
        diagram,
        reverse,
        regression,
        text_regression,
        svg_render_settings,
        text_render_settings,
    )


def check_svg_arc_radius(
    diagram,
    reverse,
    regression,
    text_regression,
    svg_render_settings,
    text_render_settings,
):
    svg_render_settings = dataclasses.replace(
        svg_render_settings,
        arc_radius=50,
        arc_margin=0,
        horizontal_seq_separation=0,
    )
    check_svg(
        diagram,
        reverse,
        regression,
        text_regression,
        svg_render_settings,
        text_render_settings,
    )


def check_svg_sep(
    diagram,
    reverse,
    regression,
    text_regression,
    svg_render_settings,
    text_render_settings,
):
    svg_render_settings = dataclasses.replace(
        svg_render_settings,
        arc_radius=0,
        arc_margin=0,
        horizontal_seq_separation=50,
    )
    check_svg(
        diagram,
        reverse,
        regression,
        text_regression,
        svg_render_settings,
        text_render_settings,
    )


def check_text(
    diagram,
    reverse,
    regression,
    text_regression,
    svg_render_settings,
    text_render_settings,
):
    rendered = rr.render_text(
        diagram,
        max_width=80,
        reverse=reverse,
        settings=text_render_settings,
    )
    text_regression(rendered)


_STACK = rr.stack(rr.terminal("A"), rr.terminal("B"))


@pytest.mark.parametrize(
    "check",
    [
        check_svg,
        # check_svg_arc_margin,
        # check_svg_arc_radius,
        # check_svg_sep,
        check_text,
    ],
    ids=[
        "svg",
        # "svg_arc_margin",
        # "svg_arc_radius",
        # "svg_sep",
        "txt",
    ],
)
@pytest.mark.parametrize(
    "reverse",
    [
        False,
        True,
    ],
    ids=[
        "ltr",
        "rtl",
    ],
)
@pytest.mark.parametrize(
    "diagram",
    [
        # Nodes
        pytest.param(
            rr.skip(),
            id="skip",
        ),
        pytest.param(
            rr.terminal("A"),
            id="terminal",
        ),
        pytest.param(
            rr.non_terminal("A"),
            id="non_terminal",
        ),
        pytest.param(
            rr.comment("A"),
            id="comment",
        ),
        pytest.param(
            rr.terminal("A\nB C\nD"),
            id="terminal_multiline",
        ),
        # Sequences
        pytest.param(
            rr.sequence(
                rr.skip(),
            ),
            id="sequence-empty",
        ),
        pytest.param(
            rr.sequence(
                rr.terminal("A"),
            ),
            id="sequence-1",
        ),
        pytest.param(
            rr.sequence(
                rr.terminal("A"),
                rr.non_terminal("B C D"),
                rr.terminal("E"),
            ),
            id="sequence-3",
        ),
        pytest.param(
            rr.stack(
                rr.terminal("A"),
                rr.non_terminal("B C D"),
                rr.terminal("E"),
            ),
            id="stack",
        ),
        pytest.param(
            rr.no_break(
                rr.terminal("A"),
                rr.non_terminal("B C D"),
                rr.terminal("E"),
            ),
            id="no_break-short",
        ),
        pytest.param(
            rr.no_break(
                rr.terminal("A very very long terminal"),
                rr.terminal("A very very long terminal"),
            ),
            id="no_break-long",
        ),
        pytest.param(
            rr.sequence(
                rr.terminal("Long terminal"),
                rr.terminal("A"),
                rr.terminal("B"),
                rr.terminal("C"),
                rr.terminal("Long terminal"),
                # rr.terminal("A very very long terminal"),
                # rr.terminal("A very very long terminal"),
                # rr.terminal("A very very long terminal"),
            ),
            id="sequence-break",
        ),
        pytest.param(
            rr.sequence(
                rr.terminal("Long terminal"),
                rr.terminal("A"),
                rr.terminal("B"),
                rr.terminal("C"),
                rr.terminal("Long terminal"),
                linebreaks=[
                    "DEFAULT",
                    "DEFAULT",
                    "SOFT",
                    "DEFAULT",
                ],
            ),
            id="sequence-soft-break",
        ),
        pytest.param(
            rr.sequence(
                rr.terminal("XX"),
                rr.choice(
                    rr.terminal("A"),
                    rr.terminal("B"),
                    rr.terminal("C"),
                    default=1,
                ),
                rr.terminal("YY"),
            ),
            id="sequence-up-down",
        ),
        pytest.param(
            rr.stack(
                rr.choice(
                    rr.terminal("A"),
                    rr.terminal("B"),
                    rr.terminal("C"),
                    default=1,
                ),
                rr.terminal("XX"),
                rr.terminal("YY"),
            ),
            id="stack-up-down-0",
        ),
        pytest.param(
            rr.stack(
                rr.terminal("XX"),
                rr.choice(
                    rr.terminal("A"),
                    rr.terminal("B"),
                    rr.terminal("C"),
                    default=1,
                ),
                rr.terminal("YY"),
            ),
            id="stack-up-down-1",
        ),
        pytest.param(
            rr.stack(
                rr.terminal("XX"),
                rr.terminal("YY"),
                rr.choice(
                    rr.terminal("A"),
                    rr.terminal("B"),
                    rr.terminal("C"),
                    default=1,
                ),
            ),
            id="stack-up-down-2",
        ),
        # Optionals
        pytest.param(
            rr.optional(
                rr.skip(),
            ),
            id="optional-empty",
        ),
        pytest.param(
            rr.optional(
                rr.terminal("A"),
            ),
            id="optional",
        ),
        pytest.param(
            rr.optional(
                rr.terminal("A"),
                skip=True,
            ),
            id="optional-skip",
        ),
        pytest.param(
            rr.optional(
                rr.terminal("A"),
                skip_bottom=True,
            ),
            id="optional-bottom",
        ),
        pytest.param(
            rr.optional(
                rr.terminal("A"),
                skip=True,
                skip_bottom=True,
            ),
            id="optional-bottom-skip",
        ),
        # Choices
        pytest.param(
            rr.choice(
                rr.skip(),
            ),
            id="choice-empty",
        ),
        pytest.param(
            rr.choice(
                rr.terminal("A"),
            ),
            id="choice-1",
        ),
        pytest.param(
            rr.choice(
                rr.terminal("A"),
                rr.terminal("B"),
                rr.terminal("C"),
            ),
            id="choice-3",
        ),
        pytest.param(
            rr.choice(
                rr.terminal("A"),
                rr.terminal("B"),
                rr.terminal("C"),
                default=1,
            ),
            id="choice-3-default-1",
        ),
        pytest.param(
            rr.choice(
                rr.terminal("A"),
                rr.terminal("B"),
                rr.terminal("C"),
                default=2,
            ),
            id="choice-3-default-2",
        ),
        pytest.param(
            rr.choice(
                rr.skip(),
                rr.terminal("A"),
                rr.skip(),
                rr.terminal("B"),
            ),
            id="choice-several-skips",
        ),
        pytest.param(
            rr.choice(
                rr.skip(),
                rr.terminal("A"),
                rr.skip(),
                rr.terminal("B"),
                default=2,
            ),
            id="choice-several-skips-default",
        ),
        pytest.param(
            rr.choice(
                rr.sequence(
                    rr.terminal("A"),
                    rr.choice(
                        rr.terminal("A"),
                        rr.terminal("B"),
                        rr.terminal("C"),
                        default=1,
                    ),
                    rr.terminal("B"),
                ),
                rr.terminal("XX"),
                rr.terminal("YY"),
            ),
            id="choice-up-down-0",
        ),
        pytest.param(
            rr.choice(
                rr.terminal("XX"),
                rr.sequence(
                    rr.terminal("A"),
                    rr.choice(
                        rr.terminal("A"),
                        rr.terminal("B"),
                        rr.terminal("C"),
                        default=1,
                    ),
                    rr.terminal("B"),
                ),
                rr.terminal("YY"),
            ),
            id="choice-up-down-1",
        ),
        pytest.param(
            rr.choice(
                rr.terminal("XX"),
                rr.terminal("YY"),
                rr.sequence(
                    rr.terminal("A"),
                    rr.choice(
                        rr.terminal("A"),
                        rr.terminal("B"),
                        rr.terminal("C"),
                        default=1,
                    ),
                    rr.terminal("B"),
                ),
                default=1,
            ),
            id="choice-up-down-2",
        ),
        pytest.param(
            rr.choice(
                rr.sequence(
                    rr.terminal("A"),
                    rr.terminal("B"),
                ),
                rr.stack(
                    rr.sequence(
                        rr.terminal("C"),
                        rr.terminal("D"),
                    ),
                    rr.terminal("E"),
                ),
                rr.terminal("F"),
                default=1,
            ),
            id="choice-up-down-complex",
        ),
        pytest.param(
            rr.choice(
                rr.sequence(
                    rr.choice(
                        rr.terminal("A"),
                        rr.terminal("B"),
                    ),
                    rr.terminal("X"),
                ),
                rr.sequence(
                    rr.terminal("XXX"),
                    rr.choice(
                        rr.terminal("A"),
                        rr.terminal("B"),
                        default=1,
                    ),
                ),
            ),
            id="choice-tucking",
        ),
        # One or more
        pytest.param(
            rr.one_or_more(
                rr.skip(),
            ),
            id="one_or_more-empty",
        ),
        pytest.param(
            rr.one_or_more(rr.skip(), repeat=rr.terminal("A")),
            id="one_or_more-non-empty-repeat",
        ),
        pytest.param(
            rr.one_or_more(
                rr.terminal("A"),
            ),
            id="one_or_more-1",
        ),
        pytest.param(
            rr.one_or_more(
                rr.terminal("A"),
                repeat=rr.terminal("B"),
            ),
            id="one_or_more-1-repeat",
        ),
        pytest.param(
            rr.one_or_more(
                rr.sequence(
                    rr.terminal("A"),
                    rr.terminal("B"),
                    rr.terminal("C"),
                ),
            ),
            id="one_or_more-3",
        ),
        pytest.param(
            rr.one_or_more(
                rr.sequence(
                    rr.terminal("A"),
                    rr.optional(rr.terminal("B")),
                    rr.terminal("C"),
                ),
                repeat=rr.sequence(
                    rr.terminal("D"),
                    rr.optional(rr.terminal("E")),
                    rr.terminal("F"),
                ),
            ),
            id="one_or_more-3-repeat",
        ),
        pytest.param(
            rr.one_or_more(
                rr.non_terminal("arg"),
                repeat=rr.choice(
                    rr.terminal(","),
                    rr.terminal(";"),
                ),
            ),
            id="one_or_more-repeat-choice",
        ),
        pytest.param(
            rr.one_or_more(
                rr.non_terminal("arg"),
                repeat=rr.sequence(
                    rr.one_or_more(
                        rr.non_terminal("arg"),
                        repeat=rr.choice(
                            rr.terminal(","),
                            rr.terminal(";"),
                        ),
                    ),
                ),
            ),
            id="one_or_more-repeat-one_or_more",
        ),
        pytest.param(
            rr.one_or_more(
                rr.stack(
                    rr.sequence(
                        rr.terminal("A"),
                        rr.terminal("B"),
                    ),
                    rr.terminal("C"),
                ),
                repeat=rr.stack(
                    rr.sequence(
                        rr.terminal("D"),
                        rr.terminal("E"),
                    ),
                    rr.terminal("F"),
                ),
            ),
            id="one_or_more-up-down",
        ),
        pytest.param(
            rr.stack(
                rr.terminal("X"),
                rr.one_or_more(
                    rr.terminal("A"),
                    repeat=rr.stack(rr.terminal("B")),
                ),
                rr.terminal("Y"),
            ),
            id="one_or_more-stack-integration",
        ),
        # Zero or more
        pytest.param(
            rr.zero_or_more(
                rr.skip(),
            ),
            id="zero_or_more-empty",
        ),
        pytest.param(
            rr.zero_or_more(rr.skip(), repeat=rr.terminal("A")),
            id="zero_or_more-non-empty-repeat",
        ),
        pytest.param(
            rr.zero_or_more(
                rr.terminal("A"),
            ),
            id="zero_or_more-1",
        ),
        pytest.param(
            rr.zero_or_more(
                rr.terminal("A"),
                repeat=rr.terminal("B"),
            ),
            id="zero_or_more-1-repeat",
        ),
        pytest.param(
            rr.zero_or_more(
                rr.sequence(
                    rr.terminal("A"),
                    rr.terminal("B"),
                    rr.terminal("C"),
                ),
            ),
            id="zero_or_more-3",
        ),
        pytest.param(
            rr.zero_or_more(
                rr.sequence(
                    rr.terminal("A"),
                    rr.optional(rr.terminal("B")),
                    rr.terminal("C"),
                ),
                repeat=rr.sequence(
                    rr.terminal("D"),
                    rr.optional(rr.terminal("E")),
                    rr.terminal("F"),
                ),
            ),
            id="zero_or_more-3-repeat",
        ),
        pytest.param(
            rr.zero_or_more(
                rr.non_terminal("arg"),
                repeat=rr.choice(
                    rr.terminal(","),
                    rr.terminal(";"),
                ),
            ),
            id="zero_or_more-repeat-choice",
        ),
        pytest.param(
            rr.zero_or_more(
                rr.non_terminal("arg"),
                repeat=rr.sequence(
                    rr.zero_or_more(
                        rr.non_terminal("arg"),
                        repeat=rr.choice(
                            rr.terminal(","),
                            rr.terminal(";"),
                        ),
                    ),
                ),
            ),
            id="zero_or_more-repeat-zero_or_more",
        ),
        pytest.param(
            rr.zero_or_more(
                rr.stack(
                    rr.sequence(
                        rr.terminal("A"),
                        rr.terminal("B"),
                    ),
                    rr.terminal("C"),
                ),
                repeat=rr.stack(
                    rr.sequence(
                        rr.terminal("D"),
                        rr.terminal("E"),
                    ),
                    rr.terminal("F"),
                ),
            ),
            id="zero_or_more-up-down",
        ),
        pytest.param(
            rr.stack(
                rr.terminal("X"),
                rr.zero_or_more(
                    rr.terminal("A"),
                    repeat=rr.stack(rr.terminal("B")),
                ),
                rr.terminal("Y"),
            ),
            id="zero_or_more-stack-integration",
        ),
        # Optimizations
        pytest.param(
            rr.sequence(
                rr.optional(
                    rr.sequence(
                        rr.optional(rr.terminal("A")),
                        rr.terminal("B"),
                    )
                ),
                rr.terminal("C"),
            ),
            id="optional-opt-start-top",
        ),
        pytest.param(
            rr.sequence(
                rr.optional(
                    rr.sequence(
                        rr.optional(rr.terminal("A"), skip_bottom=True),
                        rr.terminal("B"),
                    ),
                    skip_bottom=True,
                ),
                rr.terminal("C"),
            ),
            id="optional-opt-start-bottom",
        ),
        pytest.param(
            rr.sequence(
                rr.terminal("A"),
                rr.optional(
                    rr.sequence(
                        rr.terminal("B"),
                        rr.optional(rr.terminal("C")),
                    )
                ),
            ),
            id="optional-opt-end-top",
        ),
        pytest.param(
            rr.sequence(
                rr.terminal("A"),
                rr.optional(
                    rr.sequence(
                        rr.terminal("B"),
                        rr.optional(rr.terminal("C"), skip_bottom=True),
                    ),
                    skip_bottom=True,
                ),
            ),
            id="optional-opt-end-bottom",
        ),
        pytest.param(
            rr.choice(
                rr.sequence(
                    rr.optional(rr.terminal("A")),
                    rr.terminal("B"),
                ),
                rr.skip(),
                rr.sequence(
                    rr.optional(rr.terminal("C")),
                    rr.terminal("D"),
                ),
                default=1,
            ),
            id="optional-opt-mid-start",
        ),
        pytest.param(
            rr.choice(
                rr.sequence(
                    rr.terminal("A"),
                    rr.optional(rr.terminal("B")),
                ),
                rr.skip(),
                rr.sequence(
                    rr.terminal("C"),
                    rr.optional(rr.terminal("D")),
                ),
                default=1,
            ),
            id="optional-opt-mid-end",
        ),
        pytest.param(
            rr.choice(
                rr.sequence(
                    rr.optional(rr.terminal("A")),
                    rr.terminal("B"),
                ),
                rr.skip(),
                rr.sequence(
                    rr.terminal("C"),
                    rr.optional(rr.terminal("D")),
                ),
                default=1,
            ),
            id="optional-opt-mid-mixed",
        ),
        pytest.param(
            rr.choice(
                rr.sequence(
                    rr.optional(rr.terminal("A")),
                    rr.optional(rr.terminal("B")),
                ),
                rr.terminal("X"),
                rr.skip(),
            ),
            id="optional-opt-mid-blocked",
        ),
        pytest.param(
            rr.optional(
                rr.sequence(
                    rr.optional(rr.terminal("A")),
                    rr.optional(rr.terminal("B")),
                ),
            ),
            id="optional-opt-opt",
        ),
        pytest.param(
            rr.stack(
                rr.choice(
                    rr.terminal("A"),
                    rr.sequence(
                        rr.terminal("B"),
                        rr.choice(
                            rr.terminal("C"),
                            rr.skip(),
                            rr.terminal("D"),
                        ),
                    ),
                ),
                rr.terminal("E"),
            ),
            id="optional-opt-complex",
        ),
        pytest.param(
            yaml.safe_load(
                r"""
                    - optional:
                      - "WITH"
                      - optional: "RECURSIVE"
                      - one_or_more:
                        - non_terminal: "common-table-expression"
                        repeat: ","
                    - one_or_more:
                      - group:
                        - choice:
                          -
                            - "SELECT"
                            - choice:
                              -
                              - "DISTINCT"
                              - "ALL"
                            - one_or_more:
                              - non_terminal: "result-column"
                              repeat: ","
                            - optional:
                              - "FROM"
                              - choice:
                                - one_or_more:
                                  - non_terminal: "table-or-subquery"
                                  repeat: ","
                                - non_terminal: "join-clause"
                            - optional:
                              - "WHERE"
                              - non_terminal: "expr"
                            - optional:
                              - "GROUP"
                              - "BY"
                              - one_or_more:
                                - non_terminal: "expr"
                                repeat: ","
                            - optional:
                              - "HAVING"
                              - non_terminal: "expr"
                            - optional:
                              - "WINDOW"
                              - one_or_more:
                                - non_terminal: "window-name"
                                - "AS"
                                - non_terminal: "window-defn"
                                repeat: ","
                          -
                            - "VALUES"
                            - one_or_more:
                              - "("
                              - one_or_more:
                                  non_terminal: "expr"
                                repeat: ","
                              - ")"
                              repeat: ","
                        text: select-core
                      repeat:
                        non_terminal: "compound-operator"
                    - optional:
                      - "ORDER"
                      - "BY"
                      - one_or_more:
                        - non_terminal: "ordering-term"
                        repeat: ","
                    - optional:
                      - "LIMIT"
                      - non_terminal: "expr"
                      - choice:
                        -
                        -
                          - "OFFSET"
                          - non_terminal: "expr"
                        -
                          - ","
                          - non_terminal: "expr"
            """
            ),
            id="complex",
        ),
        pytest.param(
            rr.sequence(
                rr.terminal("A"),
                rr.group(
                    rr.choice(
                        rr.terminal("X"),
                        rr.stack(
                            rr.terminal("B"),
                            rr.terminal("C"),
                        ),
                        rr.terminal("Y"),
                        default=1,
                    )
                ),
                rr.terminal("D"),
            ),
            id="group-no-title",
        ),
        pytest.param(
            rr.sequence(
                rr.terminal("A"),
                rr.group(
                    rr.choice(
                        rr.terminal("X"),
                        rr.stack(
                            rr.terminal("B"),
                            rr.terminal("C"),
                        ),
                        rr.terminal("Y"),
                        default=1,
                    ),
                    text="Group",
                ),
                rr.terminal("D"),
            ),
            id="group-title",
        ),
        pytest.param(
            rr.sequence(
                rr.terminal("A"),
                rr.group(
                    rr.terminal("B"),
                    text="Long group title",
                ),
                rr.terminal("C"),
            ),
            id="group-long-title",
        ),
        pytest.param(
            rr.sequence(
                rr.terminal("A"),
                rr.group(
                    rr.terminal("B"),
                    text="Long\ngroup\ntitle",
                ),
                rr.terminal("C"),
            ),
            id="group-multiline-title",
        ),
    ],
)
def test_render(
    check,
    diagram,
    reverse,
    regression,
    text_regression,
    svg_render_settings,
    text_render_settings,
):
    check(
        diagram,
        reverse,
        regression,
        text_regression,
        svg_render_settings,
        text_render_settings,
    )



import io
import itertools
import os
import pathlib
from copy import deepcopy

import pytest
from pytest_image_diff.plugin import DiffInfoCallableType  # type: ignore
from reportlab.graphics import renderPM
from svglib import svglib  # type: ignore

import syntax_diagrams as rr
from syntax_diagrams._impl.render.svg import svg_layout_settings as _svg_layout_settings
from syntax_diagrams._impl.render.text import (
    text_layout_settings as _text_layout_settings,
)
from syntax_diagrams.render import DEFAULT_CSS


@pytest.fixture(scope="session")
def image_diff_root() -> str:
    return os.path.dirname(__file__)


@pytest.fixture(scope="session")  # pragma: no cover
def image_diff_dir(image_diff_root: str) -> str:
    return os.path.join(image_diff_root, "image_diff")


@pytest.fixture(scope="session")
def image_diff_reference_dir(image_diff_root: str) -> str:
    return os.path.join(image_diff_root, "image_diff_reference")


@pytest.fixture
def regression(image_regression, request):
    def regression(render, suffix=None):
        stream = io.StringIO(render)
        drawing = svglib.svg2rlg(stream)  # type: ignore
        assert drawing
        data = renderPM.drawToString(drawing, dpi=72)  # type: ignore
        assert image_regression(data, threshold=0.0001, suffix=suffix)

    return regression


@pytest.fixture
def text_regression(_image_diff_info: DiffInfoCallableType, request):
    def _factory(text: str, suffix: str | None = None) -> bool:
        diff_info = _image_diff_info("", suffix)
        reference_name = pathlib.Path(
            diff_info.reference_name.removesuffix(".png") + ".txt"
        )
        reference_name.parent.mkdir(parents=True, exist_ok=True)
        if not os.path.exists(reference_name):
            reference_name.write_text(text)
            return True
        else:
            expected = reference_name.read_text()
            if text != expected:
                ll = len(text.splitlines()[0])
                rl = len(expected.splitlines()[0])
                ml = max(ll, rl)
                lines: list[str] = []
                for l, r in itertools.zip_longest(
                    text.splitlines(), expected.splitlines()
                ):
                    diff = "".join(
                        [
                            (
                                " "
                                if (cl or " ") == (cr or " ")
                                else (cl if cl != " " else cr) or "?"
                            )
                            for cl, cr in itertools.zip_longest(
                                l or ll * " ", r or rl * " "
                            )
                        ]
                    )
                    lines.append(f"{l or ll * ' '}  :  {r or rl * ' '}  :  {diff}")
                raise AssertionError(
                    f"\n{'Got:':<{ll}}  :  {'Expected:':<{rl}}  :  {'Diff:':<{ml}}\n"
                    + "\n".join(lines)
                )
        return True

    yield _factory


@pytest.fixture
def text_layout_settings():
    return _text_layout_settings()


@pytest.fixture
def svg_layout_settings():
    return _svg_layout_settings()


@pytest.fixture
def svg_css():
    css = deepcopy(DEFAULT_CSS)
    css.setdefault("text", {})["transform"] = "translate(0, -2)"
    css.setdefault(".group text", {})["transform"] = "translate(0, -2)"
    return css


@pytest.fixture
def svg_render_settings(svg_css):
    return rr.SvgRenderSettings(css_style=svg_css)


@pytest.fixture
def text_render_settings():
    return rr.TextRenderSettings()



from __future__ import annotations

import typing as _t
from dataclasses import replace

from syntax_diagrams._impl.render import (
    LayoutContext,
    LayoutSettings,
    Render,
    RenderContext,
)
from syntax_diagrams._impl.tree import Element
from syntax_diagrams._impl.tree.barrier import Barrier
from syntax_diagrams._impl.vec import Vec

T = _t.TypeVar("T")


class Group(Element[T], _t.Generic[T]):
    _text_width: int

    def __init__(
        self,
        item: Element[T],
        text: str | None = None,
        css_class: str | None = None,
        href: str | None = None,
        title: str | None = None,
    ):
        self._text = text
        self._item = Barrier(item)
        self._css_class = css_class
        self._href = href
        self._title = title

    def _calculate_content_layout(
        self, settings: LayoutSettings[T], context: LayoutContext
    ):
        context = self._isolate()
        context = replace(
            context,
            width=max(
                0,
                context.width
                - 2 * (settings.group_vertical_padding + settings.group_thickness),
            ),
            allow_shrinking_stacks=False,
        )

        self._item.calculate_layout(settings, context)

        self._text_width, self._text_height = (
            settings.group_text_measure.measure(self._text) if self._text else (0, 0)
        )

        self.content_width = max(self._item.width, self._text_width) + 2 * (
            settings.group_horizontal_padding + settings.group_thickness
        )
        self.start_padding = 0
        self.end_padding = 0
        self.start_margin = settings.group_horizontal_margin
        self.end_margin = settings.group_horizontal_margin
        self.height = self._item.height
        self.up = (
            self._item.up
            + settings.group_vertical_padding
            + settings.group_thickness
            + self._text_height
            + (settings.group_text_vertical_offset if self._text else 0)
            + settings.group_vertical_margin
        )
        self.down = (
            self._item.down
            + settings.group_vertical_padding
            + settings.group_thickness
            + settings.group_vertical_margin
        )
        self.display_width = self.width

    def _render_content(self, render: Render[T], context: RenderContext):
        context = replace(
            context,
            pos=context.pos
            + Vec(
                context.dir
                * (
                    render.settings.group_horizontal_padding
                    + render.settings.group_thickness
                ),
                0,
            ),
        )
        if not context.reverse:
            pos = context.pos
        else:
            pos = context.pos - Vec(self.width, 0)
        render.group(
            pos
            - Vec(
                context.dir
                * (
                    render.settings.group_horizontal_padding
                    + render.settings.group_thickness
                ),
                self._item.up
                + render.settings.group_vertical_padding
                + render.settings.group_thickness,
            ),
            self.width,
            2
            * (render.settings.group_vertical_padding + render.settings.group_thickness)
            + self._item.up
            + self._item.height
            + self._item.down,
            self._css_class,
            self._text_width,
            self._text_height,
            self._text,
            self._href,
            self._title,
        )

        self._item.render(render, context)

    def __str__(self) -> str:
        return f"<{self._text}>({self._item})"
